"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""

import time
import cv2

from dlclivegui.camera import Camera, CameraError
from dlclivegui.camera.tisgrabber_windows import TIS_CAM


class TISCam(Camera):
    @staticmethod
    def arg_restrictions():

        return {"serial_number": TIS_CAM().GetDevices(), "rotate": [0, 90, 180, 270]}

    def __init__(
        self,
        serial_number="",
        resolution=[720, 540],
        exposure=0.005,
        rotate=0,
        crop=None,
        fps=100,
        display=True,
        display_resize=1.0,
    ):
        """
        Params
        ------
        serial_number = string; serial number for imaging source camera
        crop = dict; contains ints named top, left, height, width for cropping
            default = None, uses default parameters specific to camera
        """

        if (rotate == 90) or (rotate == 270):
            resolution = [resolution[1], resolution[0]]

        super().__init__(
            serial_number,
            resolution=resolution,
            exposure=exposure,
            rotate=rotate,
            crop=crop,
            fps=fps,
            use_tk_display=display,
            display_resize=display_resize,
        )
        self.display = display

    def set_exposure(self):

        val = self.exposure
        val = 1 if val > 1 else val
        val = 0 if val < 0 else val
        self.cam.SetPropertyAbsoluteValue("Exposure", "Value", val)

    def get_exposure(self):

        exposure = [0]
        self.cam.GetPropertyAbsoluteValue("Exposure", "Value", exposure)
        return round(exposure[0], 3)

    # def set_crop(self):

    #     crop = self.crop

    #     if crop:
    #         top = int(crop[0])
    #         left = int(crop[2])
    #         height = int(crop[1]-top)
    #         width = int(crop[3]-left)

    #         if not self.crop_filter:
    #             self.crop_filter = self.cam.CreateFrameFilter(b'ROI')
    #             self.cam.AddFrameFilter(self.crop_filter)

    #         self.cam.FilterSetParameter(self.crop_filter, b'Top', top)
    #         self.cam.FilterSetParameter(self.crop_filter, b'Left', left)
    #         self.cam.FilterSetParameter(self.crop_filter, b'Height', height)
    #         self.cam.FilterSetParameter(self.crop_filter, b'Width', width)

    def set_rotation(self):

        if not self.rotation_filter:
            self.rotation_filter = self.cam.CreateFrameFilter(b"Rotate Flip")
            self.cam.AddFrameFilter(self.rotation_filter)
        self.cam.FilterSetParameter(
            self.rotation_filter, b"Rotation Angle", self.rotate
        )

    def set_fps(self):

        self.cam.SetFrameRate(self.fps)

    def set_capture_device(self):

        self.cam = TIS_CAM()
        self.crop_filter = None
        self.rotation_filter = None
        self.set_rotation()
        # self.set_crop()
        self.set_fps()
        self.next_frame = time.time()

        self.cam.open(self.id)
        self.cam.SetContinuousMode(0)
        self.cam.StartLive(0)

        self.set_exposure()

        return True

    def get_image(self):

        self.cam.SnapImage()
        frame = self.cam.GetImageEx()
        frame = cv2.flip(frame, 0)
        if self.crop is not None:
            frame = frame[self.crop[0] : self.crop[1], self.crop[2] : self.crop[3]]
        return frame

    def close_capture_device(self):

        self.cam.StopLive()



from dlclivegui.camera_process import CameraProcess
from dlclivegui.pose_process import CameraPoseProcess
from dlclivegui.video import create_labeled_video
from dlclivegui.dlclivegui import DLCLiveGUI



from ..processor import Processor
import serial
import struct
import time


class TeensyLaser(Processor):
    def __init__(
        self, com, baudrate=115200, pulse_freq=50, pulse_width=5, max_stim_dur=0
    ):

        super().__init__()
        self.ser = serial.Serial(com, baudrate)
        self.pulse_freq = pulse_freq
        self.pulse_width = pulse_width
        self.max_stim_dur = (
            max_stim_dur if (max_stim_dur >= 0) and (max_stim_dur < 65356) else 0
        )
        self.stim_on = False
        self.stim_on_time = []
        self.stim_off_time = []

    def close_serial(self):

        self.ser.close()

    def stimulate_on(self):

        # command to activate PWM signal to laser is the letter 'O' followed by three 16 bit integers -- pulse frequency, pulse width, and max stim duration
        if not self.stim_on:
            self.ser.write(
                b"O"
                + struct.pack(
                    "HHH", self.pulse_freq, self.pulse_width, self.max_stim_dur
                )
            )
            self.stim_on = True
            self.stim_on_time.append(time.time())

    def stim_off(self):

        # command to turn off PWM signal to laser is the letter 'X'
        if self.stim_on:
            self.ser.write(b"X")
            self.stim_on = False
            self.stim_off_time.append(time.time())

    def process(self, pose):

        # define criteria to stimulate (e.g. if first point is in a corner of the video)
        box = [[0, 100], [0, 100]]
        if (
            (pose[0][0] > box[0][0])
            and (pose[0][0] < box[0][1])
            and (pose[0][1] > box[1][0])
            and (pose[0][1] < box[1][1])
        ):
            self.stimulate_on()
        else:
            self.stim_off()

        return pose

    def save(self, file=None):

        ### save stim on and stim off times
        save_code = 0
        if file:
            try:
                pickle.dump(
                    {"stim_on": self.stim_on_time, "stim_off": self.stim_off_time},
                    open(file, "wb"),
                )
                save_code = 1
            except Exception:
                save_code = -1
        return save_code



"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


import ctypes
import numpy as np
import time

import gi

gi.require_version("Aravis", "0.6")
from gi.repository import Aravis
import cv2

from dlclivegui.camera import Camera


class AravisCam(Camera):
    @staticmethod
    def arg_restrictions():

        Aravis.update_device_list()
        n_cams = Aravis.get_n_devices()
        ids = [Aravis.get_device_id(i) for i in range(n_cams)]
        return {"id": ids}

    def __init__(
        self,
        id="",
        resolution=[720, 540],
        exposure=0.005,
        gain=0,
        rotate=0,
        crop=None,
        fps=100,
        display=True,
        display_resize=1.0,
    ):

        super().__init__(
            id,
            resolution=resolution,
            exposure=exposure,
            gain=gain,
            rotate=rotate,
            crop=crop,
            fps=fps,
            use_tk_display=display,
            display_resize=display_resize,
        )

    def set_capture_device(self):

        self.cam = Aravis.Camera.new(self.id)
        self.no_auto()
        self.set_exposure(self.exposure)
        self.set_crop(self.crop)
        self.cam.set_frame_rate(self.fps)

        self.stream = self.cam.create_stream()
        self.stream.push_buffer(Aravis.Buffer.new_allocate(self.cam.get_payload()))
        self.cam.start_acquisition()

        return True

    def no_auto(self):

        self.cam.set_exposure_time_auto(0)
        self.cam.set_gain_auto(0)

    def set_exposure(self, val):

        val = 1 if val > 1 else val
        val = 0 if val < 0 else val
        self.cam.set_exposure_time(val * 1e6)

    def set_crop(self, crop):

        if crop:
            left = crop[0]
            width = crop[1] - left
            top = crop[3]
            height = top - crop[2]
            self.cam.set_region(left, top, width, height)
            self.im_size = (width, height)

    def get_image_on_time(self):

        buffer = None
        while buffer is None:
            buffer = self.stream.try_pop_buffer()

        frame = self._convert_image_to_numpy(buffer)
        self.stream.push_buffer(buffer)

        return frame, time.time()

    def _convert_image_to_numpy(self, buffer):
        """ from https://github.com/SintefManufacturing/python-aravis """

        pixel_format = buffer.get_image_pixel_format()
        bits_per_pixel = pixel_format >> 16 & 0xFF

        if bits_per_pixel == 8:
            INTP = ctypes.POINTER(ctypes.c_uint8)
        else:
            INTP = ctypes.POINTER(ctypes.c_uint16)

        addr = buffer.get_data()
        ptr = ctypes.cast(addr, INTP)

        frame = np.ctypeslib.as_array(
            ptr, (buffer.get_image_height(), buffer.get_image_width())
        )
        frame = frame.copy()

        if frame.ndim < 3:
            frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2RGB)

        return frame

    def close_capture_device():

        self.cam.stop_acquisition()



"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


import platform

from dlclivegui.camera.camera import Camera, CameraError
from dlclivegui.camera.opencv import OpenCVCam

if platform.system() == "Windows":
    try:
        from dlclivegui.camera.tiscamera_windows import TISCam
    except Exception as e:
        pass

if platform.system() == "Linux":
    try:
        from dlclivegui.camera.tiscamera_linux import TISCam
    except Exception as e:
        pass
        # print(f"Error importing TISCam on Linux: {e}")

if platform.system() in ["Darwin", "Linux"]:
    try:
        from dlclivegui.camera.aravis import AravisCam
    except Exception as e:
        pass
        # print(f"Error importing AravisCam: f{e}")

if platform.system() == "Darwin":
    try:
        from dlclivegui.camera.pseye import PSEyeCam
    except Exception as e:
        pass



"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


import cv2
from imutils import rotate_bound
import numpy as np
import pseyepy

from dlclivegui.camera import Camera, CameraError


class PSEyeCam(Camera):
    @staticmethod
    def arg_restrictions():

        return {
            "device": [i for i in range(pseyepy.cam_count())],
            "resolution": [[320, 240], [640, 480]],
            "fps": [30, 40, 50, 60, 75, 100, 125],
            "colour": [True, False],
            "auto_whitebalance": [True, False],
        }

    def __init__(
        self,
        device=0,
        resolution=[320, 240],
        exposure=100,
        gain=20,
        rotate=0,
        crop=None,
        fps=60,
        colour=False,
        auto_whitebalance=False,
        red_balance=125,
        blue_balance=125,
        green_balance=125,
        display=True,
        display_resize=1.0,
    ):

        super().__init__(
            device,
            resolution=resolution,
            exposure=exposure,
            gain=gain,
            rotate=rotate,
            crop=crop,
            fps=fps,
            use_tk_display=display,
            display_resize=display_resize,
        )
        self.colour = colour
        self.auto_whitebalance = auto_whitebalance
        self.red_balance = red_balance
        self.blue_balance = blue_balance
        self.green_balance = green_balance

    def set_capture_device(self):

        if self.im_size[0] == 320:
            res = pseyepy.Camera.RES_SMALL
        elif self.im_size[0] == 640:
            res = pseyepy.Camera.RES_LARGE
        else:
            raise CameraError(f"pseye resolution {self.im_size} not supported")

        self.cap = pseyepy.Camera(
            self.id,
            fps=self.fps,
            resolution=res,
            exposure=self.exposure,
            gain=self.gain,
            colour=self.colour,
            auto_whitebalance=self.auto_whitebalance,
            red_balance=self.red_balance,
            blue_balance=self.blue_balance,
            green_balance=self.green_balance,
        )

        return True

    def get_image_on_time(self):

        frame, _ = self.cap.read()

        if self.rotate != 0:
            frame = rotate_bound(frame, self.rotate)
        if self.crop:
            frame = frame[self.crop[2] : self.crop[3], self.crop[0] : self.crop[1]]

        return frame

    def close_capture_device(self):

        self.cap.end()



"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


import tkinter as tk
from tkinter import ttk
from distutils.util import strtobool


class SettingsWindow(tk.Toplevel):
    def __init__(
        self,
        title="Edit Settings",
        settings={},
        names=None,
        vals=None,
        dtypes=None,
        restrictions=None,
        parent=None,
    ):
        """ Create a tkinter settings window
        
        Parameters
        ----------
        title : str, optional
            title for window
        settings : dict, optional
            dictionary of settings with keys = setting names.
            The value for each setting should be a dictionary with three keys:
            value (a default value),
            dtype (the data type for the setting),
            restriction (a list of possible values the parameter can take on) 
        names : list, optional
            list of setting names, by default None
        vals : list, optional
            list of default values, by default None
        dtypes : list, optional
            list of setting data types, by default None
        restrictions : dict, optional
            dictionary of setting value restrictions, with keys = setting name and value = list of restrictions, by default {}
        parent : :class:`tkinter.Tk`, optional
            parent window, by default None
        
        Raises
        ------
        ValueError
            throws error if neither settings dictionary nor setting names are provided
        """

        super().__init__(parent)
        self.title(title)

        if settings:
            self.settings = settings
        elif not names:
            raise ValueError(
                "No argument names or settings dictionary. One must be provided to create a SettingsWindow."
            )
        else:
            self.settings = self.create_settings_dict(names, vals, dtypes, restrictions)

        self.cur_row = 0
        self.combobox_width = 15

        self.create_window()

    def create_settings_dict(self, names, vals=None, dtypes=None, restrictions=None):
        """Create dictionary of settings from names, vals, dtypes, and restrictions

        Parameters
        ----------
        names : list
            list of setting names
        vals : list
            list of default setting values
        dtypes : list
            list of setting dtype
        restrictions : dict
            dictionary of settting restrictions
        
        Returns
        -------
        dict
            settings dictionary with keys = names and value = dictionary with value, dtype, restrictions
        """

        set_dict = {}
        for i in range(len(names)):

            dt = dtypes[i] if dtypes is not None else None

            if vals is not None:
                val = dt(val) if type(dt) is type else [dt[0](v) for v in val]
            else:
                val = None

            restrict = restrictions[names[i]] if restrictions is not None else None

            set_dict[names[i]] = {"value": val, "dtype": dt, "restriction": restrict}

        return set_dict

    def create_window(self):
        """ Create settings GUI widgets
        """

        self.entry_vars = []
        names = tuple(self.settings.keys())
        for i in range(len(names)):

            this_setting = self.settings[names[i]]

            tk.Label(self, text=names[i] + ": ").grid(row=self.cur_row, column=0)

            v = this_setting["value"]
            if type(this_setting["dtype"]) is list:
                v = [str(x) if x is not None else "" for x in v]
                v = ", ".join(v)
            else:
                v = str(v) if v is not None else ""
            self.entry_vars.append(tk.StringVar(self, value=v))

            use_restriction = False
            if "restriction" in this_setting:
                if this_setting["restriction"] is not None:
                    use_restriction = True

            if use_restriction:
                ttk.Combobox(
                    self,
                    textvariable=self.entry_vars[-1],
                    values=this_setting["restriction"],
                    state="readonly",
                    width=self.combobox_width,
                ).grid(sticky="nsew", row=self.cur_row, column=1)
            else:
                tk.Entry(self, textvariable=self.entry_vars[-1]).grid(
                    sticky="nsew", row=self.cur_row, column=1
                )

            self.cur_row += 1

        self.cur_row += 1
        tk.Button(self, text="Update", command=self.update_vals).grid(
            sticky="nsew", row=self.cur_row, column=1
        )
        self.cur_row += 1
        tk.Button(self, text="Cancel", command=self.destroy).grid(
            sticky="nsew", row=self.cur_row, column=1
        )

        _, row_count = self.grid_size()
        for r in range(row_count):
            self.grid_rowconfigure(r, minsize=20)

    def update_vals(self):

        names = tuple(self.settings.keys())

        for i in range(len(self.entry_vars)):

            name = names[i]
            val = self.entry_vars[i].get()
            dt = (
                self.settings[name]["dtype"] if "dtype" in self.settings[name] else None
            )

            val = [v.strip() for v in val.split(",")]
            use_dt = dt if type(dt) is type else dt[0]
            use_dt = strtobool if use_dt is bool else use_dt

            try:
                val = [use_dt(v) if v else None for v in val]
            except TypeError:
                pass

            val = val if type(dt) is list else val[0]

            self.settings[name]["value"] = val

        self.quit()
        self.destroy()

    def get_values(self):

        val_dict = {}
        names = tuple(self.settings.keys())
        for i in range(len(self.settings)):
            val_dict[names[i]] = self.settings[names[i]]["value"]

        return val_dict



"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""

"""
Default processor class. Processors must contain two methods:
i) process: takes in a pose, performs operations, and returns a pose
ii) save: saves any internal data generated by the processor (such as timestamps for commands to external hardware)
"""


class Processor(object):
    def __init__(self):
        pass

    def process(self, pose):
        return pose

    def save(self, file=""):
        return 0



"""
Created on Mon Nov 21 09:44:40 2016

@author: Daniel Vassmer, Stefan_Geissler
From: https://github.com/TheImagingSource/IC-Imaging-Control-Samples/tree/master/Python

modified 10/3/2019 by Gary Kane - https://github.com/gkane26
"""

"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


from enum import Enum

import ctypes as C
import os
import sys
import numpy as np


class SinkFormats(Enum):
    Y800 = 0
    RGB24 = 1
    RGB32 = 2
    UYVY = 3
    Y16 = 4


ImageFileTypes = {"BMP": 0, "JPEG": 1}


class GrabberHandle(C.Structure):
    pass


GrabberHandle._fields_ = [("unused", C.c_int)]

##############################################################################

### GK Additions from: https://github.com/morefigs/py-ic-imaging-control


class FilterParameter(C.Structure):
    pass


FilterParameter._fields_ = [("Name", C.c_char * 30), ("Type", C.c_int)]


class FrameFilterHandle(C.Structure):
    pass


FrameFilterHandle._fields_ = [
    ("pFilter", C.c_void_p),
    ("bHasDialog", C.c_int),
    ("ParameterCount", C.c_int),
    ("Parameters", C.POINTER(FilterParameter)),
]

##############################################################################


class TIS_GrabberDLL(object):
    if sys.maxsize > 2 ** 32:
        __tisgrabber = C.windll.LoadLibrary("tisgrabber_x64.dll")
    else:
        __tisgrabber = C.windll.LoadLibrary("tisgrabber.dll")

    def __init__(self, **keyargs):
        """Initialize the Albatross from the keyword arguments."""
        self.__dict__.update(keyargs)

    GrabberHandlePtr = C.POINTER(GrabberHandle)

    ####################################

    #     Initialize the ICImagingControl class library. This function must be called
    # 	only once before any other functions of this library are called.
    # 	@param szLicenseKey IC Imaging Control license key or NULL if only a trial version is available.
    # 	@retval IC_SUCCESS on success.
    # 	@retval IC_ERROR on wrong license key or other errors.
    # 	@sa IC_CloseLibrary
    InitLibrary = __tisgrabber.IC_InitLibrary(None)

    #     Get the number of the currently available devices. This function creates an
    # 	internal array of all connected video capture devices. With each call to this
    # 	function, this array is rebuild. The name and the unique name can be retrieved
    # 	from the internal array using the functions IC_GetDevice() and IC_GetUniqueNamefromList.
    # 	They are usefull for retrieving device names for opening devices.
    #
    # 	@retval >= 0 Success, count of found devices.
    # 	@retval IC_NO_HANDLE Internal Error.
    #
    # 	@sa IC_GetDevice
    # 	@sa IC_GetUniqueNamefromList
    get_devicecount = __tisgrabber.IC_GetDeviceCount
    get_devicecount.restype = C.c_int
    get_devicecount.argtypes = None

    #     Get unique device name of a device specified by iIndex. The unique device name
    # 	consist from the device name and its serial number. It allows to differ between
    # 	more then one device of the same type connected to the computer. The unique device name
    # 	is passed to the function IC_OpenDevByUniqueName
    #
    # 	@param iIndex The number of the device whose name is to be returned. It must be
    # 				in the range from 0 to IC_GetDeviceCount(),
    # 	@return Returns the string representation of the device on success, NULL
    # 				otherwise.
    #
    # 	@sa IC_GetDeviceCount
    # 	@sa IC_GetUniqueNamefromList
    # 	@sa IC_OpenDevByUniqueName

    get_unique_name_from_list = __tisgrabber.IC_GetUniqueNamefromList
    get_unique_name_from_list.restype = C.c_char_p
    get_unique_name_from_list.argtypes = (C.c_int,)

    #     Creates a new grabber handle and returns it. A new created grabber should be
    # 	release with a call to IC_ReleaseGrabber if it is no longer needed.
    # 	@sa IC_ReleaseGrabber
    create_grabber = __tisgrabber.IC_CreateGrabber
    create_grabber.restype = GrabberHandlePtr
    create_grabber.argtypes = None

    #    Open a video capture by using its UniqueName. Use IC_GetUniqueName() to
    #    retrieve the unique name of a camera.
    #
    # 	@param hGrabber       Handle to a grabber object
    # 	@param szDisplayName  Memory that will take the display name.
    #
    # 	@sa IC_GetUniqueName
    # 	@sa IC_ReleaseGrabber
    open_device_by_unique_name = __tisgrabber.IC_OpenDevByUniqueName
    open_device_by_unique_name.restype = C.c_int
    open_device_by_unique_name.argtypes = (GrabberHandlePtr, C.c_char_p)

    set_videoformat = __tisgrabber.IC_SetVideoFormat
    set_videoformat.restype = C.c_int
    set_videoformat.argtypes = (GrabberHandlePtr, C.c_char_p)

    set_framerate = __tisgrabber.IC_SetFrameRate
    set_framerate.restype = C.c_int
    set_framerate.argtypes = (GrabberHandlePtr, C.c_float)

    #    Returns the width of the video format.
    get_video_format_width = __tisgrabber.IC_GetVideoFormatWidth
    get_video_format_width.restype = C.c_int
    get_video_format_width.argtypes = (GrabberHandlePtr,)

    #    returns the height of the video format.
    get_video_format_height = __tisgrabber.IC_GetVideoFormatHeight
    get_video_format_height.restype = C.c_int
    get_video_format_height.argtypes = (GrabberHandlePtr,)

    #    Get the number of the available video formats for the current device.
    # 	A video capture device must have been opened before this call.
    #
    # 	@param hGrabber The handle to the grabber object.
    #
    # 	@retval >= 0 Success
    # 	@retval IC_NO_DEVICE No video capture device selected.
    # 	@retval IC_NO_HANDLE No handle to the grabber object.
    #
    # 	@sa IC_GetVideoFormat
    GetVideoFormatCount = __tisgrabber.IC_GetVideoFormatCount
    GetVideoFormatCount.restype = C.c_int
    GetVideoFormatCount.argtypes = (GrabberHandlePtr,)

    #     Get a string representation of the video format specified by iIndex.
    # 	iIndex must be between 0 and IC_GetVideoFormatCount().
    # 	IC_GetVideoFormatCount() must have been called before this function,
    # 	otherwise it will always fail.
    #
    # 	@param hGrabber The handle to the grabber object.
    # 	@param iIndex Number of the video format to be used.
    #
    # 	@retval Nonnull The name of the specified video format.
    # 	@retval NULL An error occured.
    # 	@sa IC_GetVideoFormatCount
    GetVideoFormat = __tisgrabber.IC_GetVideoFormat
    GetVideoFormat.restype = C.c_char_p
    GetVideoFormat.argtypes = (GrabberHandlePtr, C.c_int)

    #    Get the number of the available input channels for the current device.
    #    A video	capture device must have been opened before this call.
    #
    # 	@param hGrabber The handle to the grabber object.
    #
    # 	@retval >= 0 Success
    # 	@retval IC_NO_DEVICE No video capture device selected.
    # 	@retval IC_NO_HANDLE No handle to the grabber object.
    #
    # 	@sa IC_GetInputChannel
    GetInputChannelCount = __tisgrabber.IC_GetInputChannelCount
    GetInputChannelCount.restype = C.c_int
    GetInputChannelCount.argtypes = (GrabberHandlePtr,)

    #     Get a string representation of the input channel specified by iIndex.
    # 	iIndex must be between 0 and IC_GetInputChannelCount().
    # 	IC_GetInputChannelCount() must have been called before this function,
    # 	otherwise it will always fail.
    # 	@param hGrabber The handle to the grabber object.
    # 	@param iIndex Number of the input channel to be used..
    #
    # 	@retval Nonnull The name of the specified input channel
    # 	@retval NULL An error occured.
    # 	@sa IC_GetInputChannelCount
    GetInputChannel = __tisgrabber.IC_GetInputChannel
    GetInputChannel.restype = C.c_char_p
    GetInputChannel.argtypes = (GrabberHandlePtr, C.c_int)

    #     Get the number of the available video norms for the current device.
    # 	A video capture device must have been opened before this call.
    #
    # 	@param hGrabber The handle to the grabber object.
    #
    # 	@retval >= 0 Success
    # 	@retval IC_NO_DEVICE No video capture device selected.
    # 	@retval IC_NO_HANDLE No handle to the grabber object.
    #
    # 	@sa IC_GetVideoNorm
    GetVideoNormCount = __tisgrabber.IC_GetVideoNormCount
    GetVideoNormCount.restype = C.c_int
    GetVideoNormCount.argtypes = (GrabberHandlePtr,)

    #     Get a string representation of the video norm specified by iIndex.
    # 	iIndex must be between 0 and IC_GetVideoNormCount().
    # 	IC_GetVideoNormCount() must have been called before this function,
    # 	otherwise it will always fail.
    #
    # 	@param hGrabber The handle to the grabber object.
    # 	@param iIndex Number of the video norm to be used.
    #
    # 	@retval Nonnull The name of the specified video norm.
    # 	@retval NULL An error occured.
    # 	@sa IC_GetVideoNormCount
    GetVideoNorm = __tisgrabber.IC_GetVideoNorm
    GetVideoNorm.restype = C.c_char_p
    GetVideoNorm.argtypes = (GrabberHandlePtr, C.c_int)

    SetFormat = __tisgrabber.IC_SetFormat
    SetFormat.restype = C.c_int
    SetFormat.argtypes = (GrabberHandlePtr, C.c_int)
    GetFormat = __tisgrabber.IC_GetFormat
    GetFormat.restype = C.c_int
    GetFormat.argtypes = (GrabberHandlePtr,)

    #    Start the live video.
    # 	@param hGrabber The handle to the grabber object.
    # 	@param iShow The parameter indicates:   @li 1 : Show the video	@li 0 : Do not show the video, but deliver frames. (For callbacks etc.)
    # 	@retval IC_SUCCESS on success
    # 	@retval IC_ERROR if something went wrong.
    # 	@sa IC_StopLive

    StartLive = __tisgrabber.IC_StartLive
    StartLive.restype = C.c_int
    StartLive.argtypes = (GrabberHandlePtr, C.c_int)

    StopLive = __tisgrabber.IC_StopLive
    StopLive.restype = C.c_int
    StopLive.argtypes = (GrabberHandlePtr,)

    SetHWND = __tisgrabber.IC_SetHWnd
    SetHWND.restype = C.c_int
    SetHWND.argtypes = (GrabberHandlePtr, C.c_int)

    #    Snaps an image. The video capture device must be set to live mode and a
    # 	sink type has to be set before this call. The format of the snapped images depend on
    # 	the selected sink type.
    #
    # 	@param hGrabber The handle to the grabber object.
    # 	@param iTimeOutMillisek The Timeout time is passed in milli seconds. A value of -1 indicates, that
    # 							no time out is set.
    #
    #
    # 	@retval IC_SUCCESS if an image has been snapped
    # 	@retval IC_ERROR if something went wrong.
    # 	@retval IC_NOT_IN_LIVEMODE if the live video has not been started.
    #
    # 	@sa IC_StartLive
    # 	@sa IC_SetFormat

    SnapImage = __tisgrabber.IC_SnapImage
    SnapImage.restype = C.c_int
    SnapImage.argtypes = (GrabberHandlePtr, C.c_int)

    #    Retrieve the properties of the current video format and sink type
    # 	@param hGrabber The handle to the grabber object.
    # 	@param *lWidth  This recieves the width of the image buffer.
    # 	@param *lHeight  This recieves the height of the image buffer.
    # 	@param *iBitsPerPixel  This recieves the count of bits per pixel.
    # 	@param *format  This recieves the current color format.
    # 	@retval IC_SUCCESS on success
    # 	@retval IC_ERROR if something went wrong.

    GetImageDescription = __tisgrabber.IC_GetImageDescription
    GetImageDescription.restype = C.c_int
    GetImageDescription.argtypes = (
        GrabberHandlePtr,
        C.POINTER(C.c_long),
        C.POINTER(C.c_long),
        C.POINTER(C.c_int),
        C.POINTER(C.c_int),
    )

    GetImagePtr = __tisgrabber.IC_GetImagePtr
    GetImagePtr.restype = C.c_void_p
    GetImagePtr.argtypes = (GrabberHandlePtr,)

    # ############################################################################
    ShowDeviceSelectionDialog = __tisgrabber.IC_ShowDeviceSelectionDialog
    ShowDeviceSelectionDialog.restype = GrabberHandlePtr
    ShowDeviceSelectionDialog.argtypes = (GrabberHandlePtr,)

    # ############################################################################

    ShowPropertyDialog = __tisgrabber.IC_ShowPropertyDialog
    ShowPropertyDialog.restype = GrabberHandlePtr
    ShowPropertyDialog.argtypes = (GrabberHandlePtr,)

    # ############################################################################
    IsDevValid = __tisgrabber.IC_IsDevValid
    IsDevValid.restype = C.c_int
    IsDevValid.argtypes = (GrabberHandlePtr,)

    # ############################################################################

    LoadDeviceStateFromFile = __tisgrabber.IC_LoadDeviceStateFromFile
    LoadDeviceStateFromFile.restype = GrabberHandlePtr
    LoadDeviceStateFromFile.argtypes = (GrabberHandlePtr, C.c_char_p)

    # ############################################################################
    SaveDeviceStateToFile = __tisgrabber.IC_SaveDeviceStateToFile
    SaveDeviceStateToFile.restype = C.c_int
    SaveDeviceStateToFile.argtypes = (GrabberHandlePtr, C.c_char_p)

    GetCameraProperty = __tisgrabber.IC_GetCameraProperty
    GetCameraProperty.restype = C.c_int
    GetCameraProperty.argtypes = (GrabberHandlePtr, C.c_int, C.POINTER(C.c_long))

    SetCameraProperty = __tisgrabber.IC_SetCameraProperty
    SetCameraProperty.restype = C.c_int
    SetCameraProperty.argtypes = (GrabberHandlePtr, C.c_int, C.c_long)

    SetPropertyValue = __tisgrabber.IC_SetPropertyValue
    SetPropertyValue.restype = C.c_int
    SetPropertyValue.argtypes = (GrabberHandlePtr, C.c_char_p, C.c_char_p, C.c_int)

    GetPropertyValue = __tisgrabber.IC_GetPropertyValue
    GetPropertyValue.restype = C.c_int
    GetPropertyValue.argtypes = (
        GrabberHandlePtr,
        C.c_char_p,
        C.c_char_p,
        C.POINTER(C.c_long),
    )

    # ############################################################################
    SetPropertySwitch = __tisgrabber.IC_SetPropertySwitch
    SetPropertySwitch.restype = C.c_int
    SetPropertySwitch.argtypes = (GrabberHandlePtr, C.c_char_p, C.c_char_p, C.c_int)

    GetPropertySwitch = __tisgrabber.IC_GetPropertySwitch
    GetPropertySwitch.restype = C.c_int
    GetPropertySwitch.argtypes = (
        GrabberHandlePtr,
        C.c_char_p,
        C.c_char_p,
        C.POINTER(C.c_long),
    )
    # ############################################################################

    IsPropertyAvailable = __tisgrabber.IC_IsPropertyAvailable
    IsPropertyAvailable.restype = C.c_int
    IsPropertyAvailable.argtypes = (GrabberHandlePtr, C.c_char_p, C.c_char_p)

    PropertyOnePush = __tisgrabber.IC_PropertyOnePush
    PropertyOnePush.restype = C.c_int
    PropertyOnePush.argtypes = (GrabberHandlePtr, C.c_char_p, C.c_char_p)

    SetPropertyAbsoluteValue = __tisgrabber.IC_SetPropertyAbsoluteValue
    SetPropertyAbsoluteValue.restype = C.c_int
    SetPropertyAbsoluteValue.argtypes = (
        GrabberHandlePtr,
        C.c_char_p,
        C.c_char_p,
        C.c_float,
    )

    GetPropertyAbsoluteValue = __tisgrabber.IC_GetPropertyAbsoluteValue
    GetPropertyAbsoluteValue.restype = C.c_int
    GetPropertyAbsoluteValue.argtypes = (
        GrabberHandlePtr,
        C.c_char_p,
        C.c_char_p,
        C.POINTER(C.c_float),
    )

    # definition of the frameready callback
    FRAMEREADYCALLBACK = C.CFUNCTYPE(
        C.c_void_p, C.c_int, C.POINTER(C.c_ubyte), C.c_ulong, C.py_object
    )

    # set callback function
    SetFrameReadyCallback = __tisgrabber.IC_SetFrameReadyCallback
    SetFrameReadyCallback.restype = C.c_int
    SetFrameReadyCallback.argtypes = [GrabberHandlePtr, FRAMEREADYCALLBACK, C.py_object]

    SetContinuousMode = __tisgrabber.IC_SetContinuousMode

    SaveImage = __tisgrabber.IC_SaveImage
    SaveImage.restype = C.c_int
    SaveImage.argtypes = [C.c_void_p, C.c_char_p, C.c_int, C.c_int]

    OpenVideoCaptureDevice = __tisgrabber.IC_OpenVideoCaptureDevice
    OpenVideoCaptureDevice.restype = C.c_int
    OpenVideoCaptureDevice.argtypes = [C.c_void_p, C.c_char_p]

    # ############################################################################

    ### GK Additions - adding frame filters. Pieces copied from: https://github.com/morefigs/py-ic-imaging-control

    CreateFrameFilter = __tisgrabber.IC_CreateFrameFilter
    CreateFrameFilter.restype = C.c_int
    CreateFrameFilter.argtypes = (C.c_char_p, C.POINTER(FrameFilterHandle))

    AddFrameFilter = __tisgrabber.IC_AddFrameFilterToDevice
    AddFrameFilter.restype = C.c_int
    AddFrameFilter.argtypes = (GrabberHandlePtr, C.POINTER(FrameFilterHandle))

    FilterGetParameter = __tisgrabber.IC_FrameFilterGetParameter
    FilterGetParameter.restype = C.c_int
    FilterGetParameter.argtypes = (C.POINTER(FrameFilterHandle), C.c_char_p, C.c_void_p)

    FilterSetParameter = __tisgrabber.IC_FrameFilterSetParameterInt
    FilterSetParameter.restype = C.c_int
    FilterSetParameter.argtypes = (C.POINTER(FrameFilterHandle), C.c_char_p, C.c_int)


# ############################################################################


class TIS_CAM(object):
    @property
    def callback_registered(self):
        return self._callback_registered

    def __init__(self):

        self._handle = C.POINTER(GrabberHandle)
        self._handle = TIS_GrabberDLL.create_grabber()
        self._callback_registered = False
        self._frame = {"num": -1, "ready": False}

    def s(self, strin):
        if sys.version[0] == "2":
            return strin
        if type(strin) == "byte":
            return strin
        return strin.encode("utf-8")

    def SetFrameReadyCallback(self, CallbackFunction, data):
        """ Set a callback function, which is called, when a new frame arrives.

            CallbackFunction : The callback function

            data : a self defined class with user data.
            """
        return TIS_GrabberDLL.SetFrameReadyCallback(
            self._handle, CallbackFunction, data
        )

    def SetContinuousMode(self, Mode):
        """ Determines, whether new frames are automatically copied into memory.

            :param Mode: If 0, all frames are copied automatically into memory. This is recommened, if the camera runs in trigger mode.
                          If 1, then snapImages must be called to get a frame into memory.
            :return: None
            """
        return TIS_GrabberDLL.SetContinuousMode(self._handle, Mode)

    def open(self, unique_device_name):
        """ Open a device

            unique_device_name : The name and serial number of the device to be opened. The device name and serial number are separated by a space.
            """
        test = TIS_GrabberDLL.open_device_by_unique_name(
            self._handle, self.s(unique_device_name)
        )

        return test

    def close(self):
        TIS_GrabberDLL.close_device(self._handle)

    def ShowDeviceSelectionDialog(self):
        self._handle = TIS_GrabberDLL.ShowDeviceSelectionDialog(self._handle)

    def ShowPropertyDialog(self):
        self._handle = TIS_GrabberDLL.ShowPropertyDialog(self._handle)

    def IsDevValid(self):
        return TIS_GrabberDLL.IsDevValid(self._handle)

    def SetHWND(self, Hwnd):
        return TIS_GrabberDLL.SetHWND(self._handle, Hwnd)

    def SaveDeviceStateToFile(self, FileName):
        return TIS_GrabberDLL.SaveDeviceStateToFile(self._handle, self.s(FileName))

    def LoadDeviceStateFromFile(self, FileName):
        self._handle = TIS_GrabberDLL.LoadDeviceStateFromFile(
            self._handle, self.s(FileName)
        )

    def SetVideoFormat(self, Format):
        return TIS_GrabberDLL.set_videoformat(self._handle, self.s(Format))

    def SetFrameRate(self, FPS):
        return TIS_GrabberDLL.set_framerate(self._handle, FPS)

    def get_video_format_width(self):
        return TIS_GrabberDLL.get_video_format_width(self._handle)

    def get_video_format_height(self):
        return TIS_GrabberDLL.get_video_format_height(self._handle)

    def GetDevices(self):
        self._Devices = []
        iDevices = TIS_GrabberDLL.get_devicecount()
        for i in range(iDevices):
            self._Devices.append(TIS_GrabberDLL.get_unique_name_from_list(i))
        return self._Devices

    def GetVideoFormats(self):
        self._Properties = []
        iVideoFormats = TIS_GrabberDLL.GetVideoFormatCount(self._handle)
        for i in range(iVideoFormats):
            self._Properties.append(TIS_GrabberDLL.GetVideoFormat(self._handle, i))
        return self._Properties

    def GetInputChannels(self):
        self.InputChannels = []
        InputChannelscount = TIS_GrabberDLL.GetInputChannelCount(self._handle)
        for i in range(InputChannelscount):
            self.InputChannels.append(TIS_GrabberDLL.GetInputChannel(self._handle, i))
        return self.InputChannels

    def GetVideoNormCount(self):
        self.GetVideoNorm = []
        GetVideoNorm_Count = TIS_GrabberDLL.GetVideoNormCount(self._handle)
        for i in range(GetVideoNorm_Count):
            self.GetVideoNorm.append(TIS_GrabberDLL.GetVideoNorm(self._handle, i))
        return self.GetVideoNorm

    def SetFormat(self, Format):
        """ SetFormat
            Sets the pixel format in memory
            @param Format Sinkformat enumeration
            """
        TIS_GrabberDLL.SetFormat(self._handle, Format.value)

    def GetFormat(self):
        val = TIS_GrabberDLL.GetFormat(self._handle)
        if val == 0:
            return SinkFormats.Y800
        if val == 2:
            return SinkFormats.RGB32
        if val == 1:
            return SinkFormats.RGB24
        if val == 3:
            return SinkFormats.UYVY
        if val == 4:
            return SinkFormats.Y16
        return SinkFormats.RGB24

    def StartLive(self, showlive=1):
        """
            Start the live video stream.

            showlive: 1 : a live video is shown, 0 : the live video is not shown.
            """
        Error = TIS_GrabberDLL.StartLive(self._handle, showlive)
        return Error

    def StopLive(self):
        """
            Stop the live video.
            """
        Error = TIS_GrabberDLL.StopLive(self._handle)
        return Error

    def SnapImage(self):
        Error = TIS_GrabberDLL.SnapImage(self._handle, 2000)
        return Error

    def GetImageDescription(self):
        lWidth = C.c_long()
        lHeight = C.c_long()
        iBitsPerPixel = C.c_int()
        COLORFORMAT = C.c_int()

        Error = TIS_GrabberDLL.GetImageDescription(
            self._handle, lWidth, lHeight, iBitsPerPixel, COLORFORMAT
        )
        return (lWidth.value, lHeight.value, iBitsPerPixel.value, COLORFORMAT.value)

    def GetImagePtr(self):
        ImagePtr = TIS_GrabberDLL.GetImagePtr(self._handle)

        return ImagePtr

    def GetImage(self):
        BildDaten = self.GetImageDescription()[:4]
        lWidth = BildDaten[0]
        lHeight = BildDaten[1]
        iBitsPerPixel = BildDaten[2] // 8

        buffer_size = lWidth * lHeight * iBitsPerPixel * C.sizeof(C.c_uint8)
        img_ptr = self.GetImagePtr()

        Bild = C.cast(img_ptr, C.POINTER(C.c_ubyte * buffer_size))

        img = np.ndarray(
            buffer=Bild.contents, dtype=np.uint8, shape=(lHeight, lWidth, iBitsPerPixel)
        )
        return img

    def GetImageEx(self):
        """ Return a numpy array with the image data tyes
            If the sink is Y16 or RGB64 (not supported yet), the dtype in the array is uint16, othereise it is uint8
            """
        BildDaten = self.GetImageDescription()[:4]
        lWidth = BildDaten[0]
        lHeight = BildDaten[1]
        iBytesPerPixel = BildDaten[2] // 8

        buffer_size = lWidth * lHeight * iBytesPerPixel * C.sizeof(C.c_uint8)
        img_ptr = self.GetImagePtr()

        Bild = C.cast(img_ptr, C.POINTER(C.c_ubyte * buffer_size))

        pixeltype = np.uint8

        if BildDaten[3] == 4:  # SinkFormats.Y16:
            pixeltype = np.uint16
            iBytesPerPixel = 1

        img = np.ndarray(
            buffer=Bild.contents,
            dtype=pixeltype,
            shape=(lHeight, lWidth, iBytesPerPixel),
        )
        return img

    def GetCameraProperty(self, iProperty):
        lFocusPos = C.c_long()
        Error = TIS_GrabberDLL.GetCameraProperty(self._handle, iProperty, lFocusPos)
        return lFocusPos.value

    def SetCameraProperty(self, iProperty, iValue):
        Error = TIS_GrabberDLL.SetCameraProperty(self._handle, iProperty, iValue)
        return Error

    def SetPropertyValue(self, Property, Element, Value):
        error = TIS_GrabberDLL.SetPropertyValue(
            self._handle, self.s(Property), self.s(Element), Value
        )
        return error

    def GetPropertyValue(self, Property, Element):
        Value = C.c_long()
        error = TIS_GrabberDLL.GetPropertyValue(
            self._handle, self.s(Property), self.s(Element), Value
        )
        return Value.value

    def PropertyAvailable(self, Property):
        Null = None
        error = TIS_GrabberDLL.IsPropertyAvailable(self._handle, self.s(Property), Null)
        return error

    def SetPropertySwitch(self, Property, Element, Value):
        error = TIS_GrabberDLL.SetPropertySwitch(
            self._handle, self.s(Property), self.s(Element), Value
        )
        return error

    def GetPropertySwitch(self, Property, Element, Value):
        lValue = C.c_long()
        error = TIS_GrabberDLL.GetPropertySwitch(
            self._handle, self.s(Property), self.s(Element), lValue
        )
        Value[0] = lValue.value
        return error

    def PropertyOnePush(self, Property, Element):
        error = TIS_GrabberDLL.PropertyOnePush(
            self._handle, self.s(Property), self.s(Element)
        )
        return error

    def SetPropertyAbsoluteValue(self, Property, Element, Value):
        error = TIS_GrabberDLL.SetPropertyAbsoluteValue(
            self._handle, self.s(Property), self.s(Element), Value
        )
        return error

    def GetPropertyAbsoluteValue(self, Property, Element, Value):
        """ Get a property value of absolute values interface, e.g. seconds or dB.
            Example code:
            ExposureTime=[0]
            Camera.GetPropertyAbsoluteValue("Exposure","Value", ExposureTime)
            print("Exposure time in secods: ", ExposureTime[0])

            :param Property: Name of the property, e.g. Gain, Exposure
            :param Element: Name of the element, e.g. "Value"
            :param Value: Object, that receives the value of the property
            :returns: 0 on success
            """
        lValue = C.c_float()
        error = TIS_GrabberDLL.GetPropertyAbsoluteValue(
            self._handle, self.s(Property), self.s(Element), lValue
        )
        Value[0] = lValue.value
        return error

    def SaveImage(self, FileName, FileType, Quality=75):
        """ Saves the last snapped image. Can by of type BMP or JPEG.
            :param FileName : Name of the mage file
            :param FileType : Determines file type, can be "JPEG" or "BMP"
            :param Quality : If file typ is JPEG, the qualitly can be given from 1 to 100.
            :return: Error code
            """
        return TIS_GrabberDLL.SaveImage(
            self._handle, self.s(FileName), IC.ImageFileTypes[self.s(FileType)], Quality
        )

    def openVideoCaptureDevice(self, DeviceName):
        """ Open the device specified by DeviceName
            :param DeviceName: Name of the device , e.g. "DFK 72AUC02"
            :returns: 1 on success, 0 otherwise.
            """
        return TIS_GrabberDLL.OpenVideoCaptureDevice(self._handle, self.s(DeviceName))

    def CreateFrameFilter(self, name):
        frame_filter_handle = FrameFilterHandle()

        err = TIS_GrabberDLL.CreateFrameFilter(
            C.c_char_p(name), C.byref(frame_filter_handle)
        )
        if err != 1:
            raise Exception("ERROR CREATING FILTER")
        return frame_filter_handle

    def AddFrameFilter(self, frame_filter_handle):
        err = TIS_GrabberDLL.AddFrameFilter(self._handle, frame_filter_handle)
        return err

    def FilterGetParameter(self, frame_filter_handle, parameter_name):
        data = C.c_int()

        err = TIS_GrabberDLL.FilterGetParameter(
            frame_filter_handle, parameter_name, C.byref(data)
        )
        return data.value

    def FilterSetParameter(self, frame_filter_handle, parameter_name, data):
        if type(data) is int:
            err = TIS_GrabberDLL.FilterSetParameter(
                frame_filter_handle, C.c_char_p(parameter_name), C.c_int(data)
            )
            return err
        else:
            raise Exception("Unknown set parameter type")



"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


import cv2
from tkinter import filedialog
from imutils import rotate_bound
import time
import platform

from dlclivegui.camera import Camera, CameraError


class OpenCVCam(Camera):
    @staticmethod
    def arg_restrictions():
        """ Returns a dictionary of arguments restrictions for DLCLiveGUI
        """

        cap = cv2.VideoCapture()
        devs = [-1]
        avail = True
        while avail:
            cur_index = devs[-1] + 1
            avail = cap.open(cur_index)
            if avail:
                devs.append(cur_index)
                cap.release()

        return {"device": devs, "display": [True, False]}

    def __init__(
        self,
        device=-1,
        file="",
        resolution=[640, 480],
        auto_exposure=0,
        exposure=0,
        gain=0,
        rotate=0,
        crop=None,
        fps=30,
        display=True,
        display_resize=1.0,
    ):

        if device != -1:
            if file:
                raise DLCLiveCameraError(
                    "A device and file were provided to OpenCVCam. Must initialize an OpenCVCam with either a device id or a video file."
                )

            self.video = False
            id = int(device)

        else:
            if not file:
                file = filedialog.askopenfilename(
                    title="Select video file for DLC-live-GUI"
                )
                if not file:
                    raise DLCLiveCameraError(
                        "Neither a device nor file were provided to OpenCVCam. Must initialize an OpenCVCam with either a device id or a video file."
                    )

            self.video = True
            cap = cv2.VideoCapture(file)
            resolution = (
                cap.get(cv2.CAP_PROP_FRAME_WIDTH),
                cap.get(cv2.CAP_PROP_FRAME_HEIGHT),
            )
            fps = cap.get(cv2.CAP_PROP_FPS)
            del cap
            id = file

        super().__init__(
            id,
            resolution=resolution,
            exposure=exposure,
            rotate=rotate,
            crop=crop,
            fps=fps,
            use_tk_display=display,
            display_resize=display_resize,
        )
        self.auto_exposure = auto_exposure
        self.gain = gain

    def set_capture_device(self):

        if not self.video:

            self.cap = (
                cv2.VideoCapture(self.id, cv2.CAP_V4L)
                if platform.system() == "Linux"
                else cv2.VideoCapture(self.id)
            )
            ret, frame = self.cap.read()

            if self.im_size:
                self.cap.set(cv2.CAP_PROP_FRAME_WIDTH, self.im_size[0])
                self.cap.set(cv2.CAP_PROP_FRAME_HEIGHT, self.im_size[1])
            if self.auto_exposure:
                self.cap.set(cv2.CAP_PROP_AUTO_EXPOSURE, self.auto_exposure)
            if self.exposure:
                self.cap.set(cv2.CAP_PROP_EXPOSURE, self.exposure)
            if self.gain:
                self.cap.set(cv2.CAP_PROP_GAIN, self.gain)
            if self.fps:
                self.cap.set(cv2.CAP_PROP_FPS, self.fps)

        else:

            self.cap = cv2.VideoCapture(self.id)

            # self.im_size = (self.cap.get(cv2.CAP_PROP_FRAME_WIDTH), self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
            # self.fps = self.cap.get(cv2.CAP_PROP_FPS)
            self.last_cap_read = 0

        self.cv2_color = self.cap.get(cv2.CAP_PROP_MODE)

        return True

    def get_image_on_time(self):

        # if video, wait...
        if self.video:
            while time.time() - self.last_cap_read < (1.0 / self.fps):
                pass

        ret, frame = self.cap.read()

        if ret:
            if self.rotate:
                frame = rotate_bound(frame, self.rotate)
            if self.crop:
                frame = frame[self.crop[2] : self.crop[3], self.crop[0] : self.crop[1]]

            if frame.ndim == 3:
                if self.cv2_color == 1:
                    frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)

            self.last_cap_read = time.time()

            return frame, self.last_cap_read
        else:
            raise CameraError("OpenCV VideoCapture.read did not return an image!")

    def close_capture_device(self):

        self.cap.release()



import multiprocess as mp
from multiprocess import queues
from queue import Queue, Empty, Full


class QueuePositionError(Exception):
    """ Error in position argument of queue read """

    pass


class ClearableQueue(Queue):
    """ A Queue that provides safe methods for writing to a full queue, reading to an empty queue, and a method to clear the queue """

    def __init__(self, maxsize=0):

        super().__init__(maxsize)

    def clear(self):
        """ Clears queue, returns all objects in a list

        Returns
        -------
        list
            list of objects from the queue
        """

        objs = []

        try:
            while True:
                objs.append(self.get_nowait())
        except Empty:
            pass

        return objs

    def write(self, obj, clear=False):
        """ Puts an object in the queue, with an option to clear queue before writing.

        Parameters
        ----------     
        obj : [type]
            An object to put in the queue
        clear : bool, optional
            flag to clear queue before putting, by default False
        
        Returns
        -------
        bool
            if write was sucessful, returns True
        """

        if clear:
            self.clear()

        try:
            self.put_nowait(obj)
            success = True
        except Full:
            success = False

        return success

    def read(self, clear=False, position="last"):
        """ Gets an object in the queue, with the option to clear the queue and return the first element, last element, or all elements
        
        Parameters
        ----------
        clear : bool, optional
            flag to clear queue before putting, by default False
        position : str, optional
            If clear is True, returned object depends on position. 
            If position = "last", returns last object. 
            If position = "first", returns first object. 
            If position = "all", returns all objects from the queue.
        
        Returns
        -------
        object
            object retrieved from the queue
        """

        obj = None

        if clear:

            objs = self.clear()

            if len(objs) > 0:
                if position == "first":
                    obj = objs[0]
                elif position == "last":
                    obj = objs[-1]
                elif position == "all":
                    obj = objs
                else:
                    raise QueuePositionError(
                        "Queue read position should be one of 'first', 'last', or 'all'"
                    )
        else:

            try:
                obj = self.get_nowait()
            except Empty:
                pass

        return obj


class ClearableMPQueue(mp.queues.Queue):
    """ A multiprocess Queue that provides safe methods for writing to a full queue, reading to an empty queue, and a method to clear the queue """

    def __init__(self, maxsize=0, ctx=mp.get_context("spawn")):

        super().__init__(maxsize, ctx=ctx)

    def clear(self):
        """ Clears queue, returns all objects in a list

        Returns
        -------
        list
            list of objects from the queue
        """

        objs = []

        try:
            while True:
                objs.append(self.get_nowait())
        except Empty:
            pass

        return objs

    def write(self, obj, clear=False):
        """ Puts an object in the queue, with an option to clear queue before writing.

        Parameters
        ----------     
        obj : [type]
            An object to put in the queue
        clear : bool, optional
            flag to clear queue before putting, by default False
        
        Returns
        -------
        bool
            if write was sucessful, returns True
        """

        if clear:
            self.clear()

        try:
            self.put_nowait(obj)
            success = True
        except Full:
            success = False

        return success

    def read(self, clear=False, position="last"):
        """ Gets an object in the queue, with the option to clear the queue and return the first element, last element, or all elements
        
        Parameters
        ----------
        clear : bool, optional
            flag to clear queue before putting, by default False
        position : str, optional
            If clear is True, returned object depends on position. 
            If position = "last", returns last object. 
            If position = "first", returns first object. 
            If position = "all", returns all objects from the queue.
        
        Returns
        -------
        object
            object retrieved from the queue
        """

        obj = None

        if clear:

            objs = self.clear()

            if len(objs) > 0:
                if position == "first":
                    obj = objs[0]
                elif position == "last":
                    obj = objs[-1]
                elif position == "all":
                    obj = objs
                else:
                    raise QueuePositionError(
                        "Queue read position should be one of 'first', 'last', or 'all'"
                    )

        else:

            try:
                obj = self.get_nowait()
            except Empty:
                pass

        return obj



"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


import os
import numpy as np
import pandas as pd
import cv2
import colorcet as cc
from PIL import ImageColor
from tqdm import tqdm


def create_labeled_video(
    data_dir,
    out_dir=None,
    dlc_online=True,
    save_images=False,
    cut=(0, np.Inf),
    crop=None,
    cmap="bmy",
    radius=3,
    lik_thresh=0.5,
    write_ts=False,
    write_scale=2,
    write_pos="bottom-left",
    write_ts_offset=0,
    display=False,
    progress=True,
    label=True,
):
    """ Create a labeled video from DeepLabCut-live-GUI recording

    Parameters
    ----------
    data_dir : str
        path to data directory
    dlc_online : bool, optional
        flag indicating dlc keypoints from online tracking, using DeepLabCut-live-GUI, or offline tracking, using :func:`dlclive.benchmark_videos`
    out_file : str, optional
        path for output file. If None, output file will be "'video_file'_LABELED.avi". by default None. If NOn
    save_images : bool, optional
        boolean flag to save still images in a folder
    cut : tuple, optional
        time of video to use. Will only save labeled video for time after cut[0] and before cut[1], by default (0, np.Inf)
    cmap : str, optional
        a :package:`colorcet` colormap, by default 'bmy'
    radius : int, optional
        radius for keypoints, by default 3
    lik_thresh : float, optional
        likelihood threshold to plot keypoints, by default 0.5
    display : bool, optional
        boolean flag to display images as video is written, by default False
    progress : bool, optional
        boolean flag to display progress bar

    Raises
    ------
    Exception
        if frames cannot be read from the video file
    """

    base_dir = os.path.basename(data_dir)
    video_file = os.path.normpath(f"{data_dir}/{base_dir}_VIDEO.avi")
    ts_file = os.path.normpath(f"{data_dir}/{base_dir}_TS.npy")
    dlc_file = (
        os.path.normpath(f"{data_dir}/{base_dir}_DLC.hdf5")
        if dlc_online
        else os.path.normpath(f"{data_dir}/{base_dir}_VIDEO_DLCLIVE_POSES.h5")
    )

    cap = cv2.VideoCapture(video_file)
    cam_frame_times = np.load(ts_file)
    n_frames = cam_frame_times.size

    lab = "LABELED" if label else "UNLABELED"
    if out_dir:
        out_file = (
            f"{out_dir}/{os.path.splitext(os.path.basename(video_file))[0]}_{lab}.avi"
        )
        out_times_file = (
            f"{out_dir}/{os.path.splitext(os.path.basename(ts_file))[0]}_{lab}.npy"
        )
    else:
        out_file = f"{os.path.splitext(video_file)[0]}_{lab}.avi"
        out_times_file = f"{os.path.splitext(ts_file)[0]}_{lab}.npy"

    os.makedirs(os.path.normpath(os.path.dirname(out_file)), exist_ok=True)

    if save_images:
        im_dir = os.path.splitext(out_file)[0]
        os.makedirs(im_dir, exist_ok=True)

    im_size = (
        int(cap.get(cv2.CAP_PROP_FRAME_WIDTH)),
        int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT)),
    )
    if crop is not None:
        crop[0] = crop[0] if crop[0] > 0 else 0
        crop[1] = crop[1] if crop[1] > 0 else im_size[1]
        crop[2] = crop[2] if crop[2] > 0 else 0
        crop[3] = crop[3] if crop[3] > 0 else im_size[0]
        im_size = (crop[3] - crop[2], crop[1] - crop[0])

    fourcc = cv2.VideoWriter_fourcc(*"DIVX")
    fps = cap.get(cv2.CAP_PROP_FPS)
    vwriter = cv2.VideoWriter(out_file, fourcc, fps, im_size)
    label_times = []

    if write_ts:
        ts_font = cv2.FONT_HERSHEY_PLAIN

        if "left" in write_pos:
            ts_w = 0
        else:
            ts_w = (
                im_size[0] if crop is None else (crop[3] - crop[2]) - (55 * write_scale)
            )

        if "bottom" in write_pos:
            ts_h = im_size[1] if crop is None else (crop[1] - crop[0])
        else:
            ts_h = 0 if crop is None else crop[0] + (12 * write_scale)

        ts_coord = (ts_w, ts_h)
        ts_color = (255, 255, 255)
        ts_size = 2

    poses = pd.read_hdf(dlc_file)
    if dlc_online:
        pose_times = poses["pose_time"]
    else:
        poses["frame_time"] = cam_frame_times
        poses["pose_time"] = cam_frame_times
    poses = poses.melt(id_vars=["frame_time", "pose_time"])
    bodyparts = poses["bodyparts"].unique()

    all_colors = getattr(cc, cmap)
    colors = [
        ImageColor.getcolor(c, "RGB")[::-1]
        for c in all_colors[:: int(len(all_colors) / bodyparts.size)]
    ]

    ind = 0
    vid_time = 0
    while vid_time < cut[0]:

        cur_time = cam_frame_times[ind]
        vid_time = cur_time - cam_frame_times[0]
        ret, frame = cap.read()
        ind += 1

        if not ret:
            raise Exception(
                f"Could not read frame = {ind+1} at time = {cur_time-cam_frame_times[0]}."
            )

    frame_times_sub = cam_frame_times[
        (cam_frame_times - cam_frame_times[0] > cut[0])
        & (cam_frame_times - cam_frame_times[0] < cut[1])
    ]
    iterator = (
        tqdm(range(ind, ind + frame_times_sub.size))
        if progress
        else range(ind, ind + frame_times_sub.size)
    )
    this_pose = np.zeros((bodyparts.size, 3))

    for i in iterator:

        cur_time = cam_frame_times[i]
        vid_time = cur_time - cam_frame_times[0]
        ret, frame = cap.read()

        if not ret:
            raise Exception(
                f"Could not read frame = {i+1} at time = {cur_time-cam_frame_times[0]}."
            )

        if dlc_online:
            poses_before_index = np.where(pose_times < cur_time)[0]
            if poses_before_index.size > 0:
                cur_pose_time = pose_times[poses_before_index[-1]]
                this_pose = poses[poses["pose_time"] == cur_pose_time]
        else:
            this_pose = poses[poses["frame_time"] == cur_time]

        if label:
            for j in range(bodyparts.size):
                this_bp = this_pose[this_pose["bodyparts"] == bodyparts[j]][
                    "value"
                ].values
                if this_bp[2] > lik_thresh:
                    x = int(this_bp[0])
                    y = int(this_bp[1])
                    frame = cv2.circle(frame, (x, y), radius, colors[j], thickness=-1)

        if crop is not None:
            frame = frame[crop[0] : crop[1], crop[2] : crop[3]]

        if write_ts:
            frame = cv2.putText(
                frame,
                f"{(vid_time-write_ts_offset):0.3f}",
                ts_coord,
                ts_font,
                write_scale,
                ts_color,
                ts_size,
            )

        if display:
            cv2.imshow("DLC Live Labeled Video", frame)
            cv2.waitKey(1)

        vwriter.write(frame)
        label_times.append(cur_time)
        if save_images:
            new_file = f"{im_dir}/frame_{i}.png"
            cv2.imwrite(new_file, frame)

    if display:
        cv2.destroyAllWindows()

    vwriter.release()
    np.save(out_times_file, label_times)


def main():

    import argparse
    import os

    parser = argparse.ArgumentParser()
    parser.add_argument("dir", type=str)
    parser.add_argument("-o", "--out-dir", type=str, default=None)
    parser.add_argument("--dlc-offline", action="store_true")
    parser.add_argument("-s", "--save-images", action="store_true")
    parser.add_argument("-u", "--cut", nargs="+", type=float, default=[0, np.Inf])
    parser.add_argument("-c", "--crop", nargs="+", type=int, default=None)
    parser.add_argument("-m", "--cmap", type=str, default="bmy")
    parser.add_argument("-r", "--radius", type=int, default=3)
    parser.add_argument("-l", "--lik-thresh", type=float, default=0.5)
    parser.add_argument("-w", "--write-ts", action="store_true")
    parser.add_argument("--write-scale", type=int, default=2)
    parser.add_argument("--write-pos", type=str, default="bottom-left")
    parser.add_argument("--write-ts-offset", type=float, default=0.0)
    parser.add_argument("-d", "--display", action="store_true")
    parser.add_argument("--no-progress", action="store_false")
    parser.add_argument("--no-label", action="store_false")
    args = parser.parse_args()

    create_labeled_video(
        args.dir,
        out_dir=args.out_dir,
        dlc_online=(not args.dlc_offline),
        save_images=args.save_images,
        cut=tuple(args.cut),
        crop=args.crop,
        cmap=args.cmap,
        radius=args.radius,
        lik_thresh=args.lik_thresh,
        write_ts=args.write_ts,
        write_scale=args.write_scale,
        write_pos=args.write_pos,
        write_ts_offset=args.write_ts_offset,
        display=args.display,
        progress=args.no_progress,
        label=args.no_label,
    )



"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


import warnings
import numpy as np
import time

import gi

gi.require_version("Tcam", "0.1")
gi.require_version("Gst", "1.0")
from gi.repository import Tcam, Gst, GLib, GObject

from dlclivegui.camera import Camera


class TISCam(Camera):

    FRAME_RATE_OPTIONS = [15, 30, 60, 120, 240, 480]
    FRAME_RATE_FRACTIONS = ["15/1", "30/1", "60/1", "120/1", "5000000/20833", "480/1"]
    IM_FORMAT = (720, 540)
    ROTATE_OPTIONS = ["identity", "90r", "180", "90l", "horiz", "vert"]

    @staticmethod
    def arg_restrictions():

        if not Gst.is_initialized():
            Gst.init()

        source = Gst.ElementFactory.make("tcambin")
        return {
            "serial_number": source.get_device_serials(),
            "fps": TISCam.FRAME_RATE_OPTIONS,
            "rotate": TISCam.ROTATE_OPTIONS,
            "color": [True, False],
            "display": [True, False],
        }

    def __init__(
        self,
        serial_number="",
        resolution=[720, 540],
        exposure=0.005,
        rotate="identity",
        crop=None,
        fps=120,
        color=False,
        display=True,
        tk_resize=1.0,
    ):

        super().__init__(
            serial_number,
            resolution=resolution,
            exposure=exposure,
            rotate=rotate,
            crop=crop,
            fps=fps,
            use_tk_display=(not display),
            display_resize=tk_resize,
        )
        self.color = color
        self.display = display
        self.sample_locked = False
        self.new_sample = False

    def no_auto(self):

        self.cam.set_tcam_property("Exposure Auto", GObject.Value(bool, False))

    def set_exposure(self, val):

        val = 1 if val > 1 else val
        val = 0 if val < 0 else val
        self.cam.set_tcam_property("Exposure", val * 1e6)

    def set_crop(self, crop):

        if crop:
            self.gst_crop = self.gst_pipeline.get_by_name("crop")
            self.gst_crop.set_property("left", crop[0])
            self.gst_crop.set_property("right", TISCam.IM_FORMAT[0] - crop[1])
            self.gst_crop.set_property("top", crop[2])
            self.gst_crop.set_property("bottom", TISCam.IM_FORMAT[1] - crop[3])
            self.im_size = (crop[3] - crop[2], crop[1] - crop[0])

    def set_rotation(self, val):

        if val:
            self.gst_rotate = self.gst_pipeline.get_by_name("rotate")
            self.gst_rotate.set_property("video-direction", val)

    def set_sink(self):

        self.gst_sink = self.gst_pipeline.get_by_name("sink")
        self.gst_sink.set_property("max-buffers", 1)
        self.gst_sink.set_property("drop", 1)
        self.gst_sink.set_property("emit-signals", True)
        self.gst_sink.connect("new-sample", self.get_image)

    def setup_gst(self, serial_number, fps):

        if not Gst.is_initialized():
            Gst.init()

        fps_index = np.where(
            [int(fps) == int(opt) for opt in TISCam.FRAME_RATE_OPTIONS]
        )[0][0]
        fps_frac = TISCam.FRAME_RATE_FRACTIONS[fps_index]
        fmat = "BGRx" if self.color else "GRAY8"

        pipeline = (
            "tcambin name=cam "
            "! videocrop name=crop "
            "! videoflip name=rotate "
            "! video/x-raw,format={},framerate={} ".format(fmat, fps_frac)
        )

        if self.display:
            pipe_sink = (
                "! tee name=t "
                "t. ! queue ! videoconvert ! ximagesink "
                "t. ! queue ! appsink name=sink"
            )
        else:
            pipe_sink = "! appsink name=sink"

        pipeline += pipe_sink

        self.gst_pipeline = Gst.parse_launch(pipeline)

        self.cam = self.gst_pipeline.get_by_name("cam")
        self.cam.set_property("serial", serial_number)

        self.set_exposure(self.exposure)
        self.set_crop(self.crop)
        self.set_rotation(self.rotate)
        self.set_sink()

    def set_capture_device(self):

        self.setup_gst(self.id, self.fps)
        self.gst_pipeline.set_state(Gst.State.PLAYING)

        return True

    def get_image(self, sink):

        # wait for sample to unlock
        while self.sample_locked:
            pass

        try:

            self.sample = sink.get_property("last-sample")
            self._convert_image_to_numpy()

        except GLib.Error as e:

            warnings.warn("Error reading image :: {}".format(e))

        finally:

            return 0

    def _convert_image_to_numpy(self):

        self.sample_locked = True

        buffer = self.sample.get_buffer()
        struct = self.sample.get_caps().get_structure(0)

        height = struct.get_value("height")
        width = struct.get_value("width")
        fmat = struct.get_value("format")
        dtype = np.uint16 if fmat == "GRAY16_LE" else np.uint8
        ncolors = 1 if "GRAY" in fmat else 4

        self.frame = np.ndarray(
            shape=(height, width, ncolors),
            buffer=buffer.extract_dup(0, buffer.get_size()),
            dtype=dtype,
        )

        self.sample_locked = False
        self.new_sample = True

    def get_image_on_time(self):

        # wait for new sample
        while not self.new_sample:
            pass
        self.new_sample = False

        return self.frame, time.time()

    def close_capture_device(self):

        self.gst_pipeline.set_state(Gst.State.NULL)



"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="deeplabcut-live-gui",
    version="1.0",
    author="A. & M. Mathis Labs",
    author_email="adim@deeplabcut.org",
    description="GUI to run real time deeplabcut experiments",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DeepLabCut/DeepLabCut-live-GUI",
    python_requires=">=3.5, <3.8",
    install_requires=[
        "deeplabcut-live",
        "pyserial",
        "pandas",
        "tables",
        "multiprocess",
        "imutils",
        "pillow",
        "tqdm",
    ],
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Operating System :: OS Independent",
    ),
    entry_points={
        "console_scripts": [
            "dlclivegui=dlclivegui.dlclivegui:main",
            "dlclivegui-video=dlclivegui.video:main",
        ]
    },
)



"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


import time
import multiprocess as mp
import ctypes
from dlclivegui.queue import ClearableQueue, ClearableMPQueue
import threading
import cv2
import numpy as np
import os


class CameraProcessError(Exception):
    """
    Exception for incorrect use of Cameras
    """

    pass


class CameraProcess(object):
    """ Camera Process Manager class. Controls image capture and writing images to a video file in a background process. 
    
    Parameters
    ----------
    device : :class:`cameracontrol.Camera`
        a camera object
    ctx : :class:`multiprocess.Context`
        multiprocessing context
    """

    def __init__(self, device, ctx=mp.get_context("spawn")):
        """ Constructor method
        """

        self.device = device
        self.ctx = ctx

        res = self.device.im_size
        self.frame_shared = mp.Array(ctypes.c_uint8, res[1] * res[0] * 3)
        self.frame = np.frombuffer(self.frame_shared.get_obj(), dtype="uint8").reshape(
            res[1], res[0], 3
        )
        self.frame_time_shared = mp.Array(ctypes.c_double, 1)
        self.frame_time = np.frombuffer(self.frame_time_shared.get_obj(), dtype="d")

        self.q_to_process = ClearableMPQueue(ctx=self.ctx)
        self.q_from_process = ClearableMPQueue(ctx=self.ctx)
        self.write_frame_queue = ClearableMPQueue(ctx=self.ctx)

        self.capture_process = None
        self.writer_process = None

    def start_capture_process(self, timeout=60):

        cmds = self.q_to_process.read(clear=True, position="all")
        if cmds is not None:
            for c in cmds:
                if c[1] != "capture":
                    self.q_to_process.write(c)

        self.capture_process = self.ctx.Process(
            target=self._run_capture,
            args=(self.frame_shared, self.frame_time_shared),
            daemon=True,
        )
        self.capture_process.start()

        stime = time.time()
        while time.time() - stime < timeout:
            cmd = self.q_from_process.read()
            if cmd is not None:
                if (cmd[0] == "capture") and (cmd[1] == "start"):
                    return cmd[2]
                else:
                    self.q_to_process.write(cmd)

        return True

    def _run_capture(self, frame_shared, frame_time):

        res = self.device.im_size
        self.frame = np.frombuffer(frame_shared.get_obj(), dtype="uint8").reshape(
            res[1], res[0], 3
        )
        self.frame_time = np.frombuffer(frame_time.get_obj(), dtype="d")

        ret = self.device.set_capture_device()
        if not ret:
            raise CameraProcessError("Could not start capture device.")
        self.q_from_process.write(("capture", "start", ret))

        self._capture_loop()

        self.device.close_capture_device()
        self.q_from_process.write(("capture", "end", True))

    def _capture_loop(self):
        """ Acquires frames from frame capture device in a loop
        """

        run = True
        write = False
        last_frame_time = time.time()

        while run:

            start_capture = time.time()

            frame, frame_time = self.device.get_image_on_time()

            write_capture = time.time()

            np.copyto(self.frame, frame)
            self.frame_time[0] = frame_time

            if write:
                ret = self.write_frame_queue.write((frame, frame_time))

            end_capture = time.time()

            # print("read frame = %0.6f // write to queues = %0.6f" % (write_capture-start_capture, end_capture-write_capture))
            # print("capture rate = %d" % (int(1 / (time.time()-last_frame_time))))
            # print("\n")

            last_frame_time = time.time()

            ### read commands
            cmd = self.q_to_process.read()
            if cmd is not None:
                if cmd[0] == "capture":
                    if cmd[1] == "write":
                        write = cmd[2]
                        self.q_from_process.write(cmd)
                    elif cmd[1] == "end":
                        run = False
                else:
                    self.q_to_process.write(cmd)

    def stop_capture_process(self):

        ret = True
        if self.capture_process is not None:
            if self.capture_process.is_alive():
                self.q_to_process.write(("capture", "end"))

                while True:
                    cmd = self.q_from_process.read()
                    if cmd is not None:
                        if cmd[0] == "capture":
                            if cmd[1] == "end":
                                break
                        else:
                            self.q_from_process.write(cmd)

                self.capture_process.join(5)
                if self.capture_process.is_alive():
                    self.capture_process.terminate()

        return True

    def start_writer_process(self, filename, timeout=60):

        cmds = self.q_to_process.read(clear=True, position="all")
        if cmds is not None:
            for c in cmds:
                if c[1] != "writer":
                    self.q_to_process.write(c)

        self.writer_process = self.ctx.Process(
            target=self._run_writer, args=(filename,), daemon=True
        )
        self.writer_process.start()

        stime = time.time()
        while time.time() - stime < timeout:
            cmd = self.q_from_process.read()
            if cmd is not None:
                if (cmd[0] == "writer") and (cmd[1] == "start"):
                    return cmd[2]
                else:
                    self.q_to_process.write(cmd)

        return True

    def _run_writer(self, filename):

        ret = self._create_writer(filename)
        self.q_from_process.write(("writer", "start", ret))

        save = self._write_loop()

        ret = self._save_video(not save)
        self.q_from_process.write(("writer", "end", ret))

    def _create_writer(self, filename):

        self.filename = filename
        self.video_file = f"{self.filename}_VIDEO.avi"
        self.timestamp_file = f"{self.filename}_TS.npy"

        self.video_writer = cv2.VideoWriter(
            self.video_file,
            cv2.VideoWriter_fourcc(*"DIVX"),
            self.device.fps,
            self.device.im_size,
        )
        self.write_frame_ts = []

        return True

    def _write_loop(self):
        """ read frames from write_frame_queue and write to file
        """

        run = True
        new_frame = None

        while run or (new_frame is not None):

            new_frame = self.write_frame_queue.read()
            if new_frame is not None:
                frame, ts = new_frame
                if frame.shape[2] == 1:
                    frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2RGB)
                self.video_writer.write(frame)
                self.write_frame_ts.append(ts)

            cmd = self.q_to_process.read()
            if cmd is not None:
                if cmd[0] == "writer":
                    if cmd[1] == "end":
                        run = False
                        save = cmd[2]
                else:
                    self.q_to_process.write(cmd)

        return save

    def _save_video(self, delete=False):

        ret = False

        self.video_writer.release()

        if (not delete) and (len(self.write_frame_ts) > 0):
            np.save(self.timestamp_file, self.write_frame_ts)
            ret = True
        else:
            os.remove(self.video_file)
            if os.path.isfile(self.timestamp_file):
                os.remove(self.timestamp_file)

        return ret

    def stop_writer_process(self, save=True):

        ret = False
        if self.writer_process is not None:
            if self.writer_process.is_alive():
                self.q_to_process.write(("writer", "end", save))

                while True:
                    cmd = self.q_from_process.read()
                    if cmd is not None:
                        if cmd[0] == "writer":
                            if cmd[1] == "end":
                                ret = cmd[2]
                                break
                        else:
                            self.q_from_process.write(cmd)

                self.writer_process.join(5)
                if self.writer_process.is_alive():
                    self.writer_process.terminate()

        return ret

    def start_record(self, timeout=5):

        ret = False

        if (self.capture_process is not None) and (self.writer_process is not None):
            if self.capture_process.is_alive() and self.writer_process.is_alive():
                self.q_to_process.write(("capture", "write", True))

                stime = time.time()
                while time.time() - stime < timeout:
                    cmd = self.q_from_process.read()
                    if cmd is not None:
                        if (cmd[0] == "capture") and (cmd[1] == "write"):
                            ret = cmd[2]
                            break
                        else:
                            self.q_from_process.write(cmd)

        return ret

    def stop_record(self, timeout=5):

        ret = False

        if (self.capture_process is not None) and (self.writer_process is not None):
            if (self.capture_process.is_alive()) and (self.writer_process.is_alive()):
                self.q_to_process.write(("capture", "write", False))

                stime = time.time()
                while time.time() - stime < timeout:
                    cmd = self.q_from_process.read()
                    if cmd is not None:
                        if (cmd[0] == "capture") and (cmd[1] == "write"):
                            ret = not cmd[2]
                            break
                        else:
                            self.q_from_process.write(cmd)

        return ret

    def get_display_frame(self):

        frame = self.frame.copy()
        if frame is not None:
            if self.device.display_resize != 1:
                frame = cv2.resize(
                    frame,
                    (
                        int(frame.shape[1] * self.device.display_resize),
                        int(frame.shape[0] * self.device.display_resize),
                    ),
                )

        return frame



"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


from tkinter import (
    Tk,
    Toplevel,
    Label,
    Entry,
    Button,
    Radiobutton,
    Checkbutton,
    StringVar,
    IntVar,
    BooleanVar,
    filedialog,
    messagebox,
    simpledialog,
)
from tkinter.ttk import Combobox
import os
import sys
import glob
import json
import datetime
import inspect
import importlib

from PIL import Image, ImageTk, ImageDraw
import colorcet as cc

from dlclivegui import CameraPoseProcess
from dlclivegui import processor
from dlclivegui import camera
from dlclivegui.tkutil import SettingsWindow


class DLCLiveGUI(object):
    """ GUI to run DLC Live experiment
    """

    def __init__(self):
        """ Constructor method
        """

        ### check if documents path exists

        if not os.path.isdir(self.get_docs_path()):
            os.mkdir(self.get_docs_path())
        if not os.path.isdir(os.path.dirname(self.get_config_path(""))):
            os.mkdir(os.path.dirname(self.get_config_path("")))

        ### get configuration ###

        self.cfg_list = [
            os.path.splitext(os.path.basename(f))[0]
            for f in glob.glob(os.path.dirname(self.get_config_path("")) + "/*.json")
        ]

        ### initialize variables

        self.cam_pose_proc = None
        self.dlc_proc_params = None

        self.display_window = None
        self.display_cmap = None
        self.display_colors = None
        self.display_radius = None
        self.display_lik_thresh = None

        ### create GUI window ###

        self.createGUI()

    def get_docs_path(self):
        """ Get path to documents folder

        Returns
        -------
        str
            path to documents folder
        """

        return os.path.normpath(os.path.expanduser("~/Documents/DeepLabCut-live-GUI"))

    def get_config_path(self, cfg_name):
        """ Get path to configuration foler

        Parameters
        ----------
        cfg_name : str
            name of config file

        Returns
        -------
        str
            path to configuration file
        """

        return os.path.normpath(self.get_docs_path() + "/config/" + cfg_name + ".json")

    def get_config(self, cfg_name):
        """ Read configuration

        Parameters
        ----------
        cfg_name : str
            name of configuration
        """

        ### read configuration file ###

        self.cfg_file = self.get_config_path(cfg_name)
        if os.path.isfile(self.cfg_file):
            cfg = json.load(open(self.cfg_file))
        else:
            cfg = {}

        ### check config ###

        cfg["cameras"] = {} if "cameras" not in cfg else cfg["cameras"]
        cfg["processor_dir"] = (
            [] if "processor_dir" not in cfg else cfg["processor_dir"]
        )
        cfg["processor_args"] = (
            {} if "processor_args" not in cfg else cfg["processor_args"]
        )
        cfg["dlc_options"] = {} if "dlc_options" not in cfg else cfg["dlc_options"]
        cfg["dlc_display_options"] = (
            {} if "dlc_display_options" not in cfg else cfg["dlc_display_options"]
        )
        cfg["subjects"] = [] if "subjects" not in cfg else cfg["subjects"]
        cfg["directories"] = [] if "directories" not in cfg else cfg["directories"]

        self.cfg = cfg

    def change_config(self, event=None):
        """ Change configuration, update GUI menus

        Parameters
        ----------
        event : tkinter event, optional
            event , by default None
        """

        if self.cfg_name.get() == "Create New Config":
            new_name = simpledialog.askstring(
                "", "Please enter a name (no special characters).", parent=self.window
            )
            self.cfg_name.set(new_name)
        self.get_config(self.cfg_name.get())

        self.camera_entry["values"] = tuple(self.cfg["cameras"].keys()) + (
            "Add Camera",
        )
        self.camera_name.set("")
        self.dlc_proc_dir_entry["values"] = tuple(self.cfg["processor_dir"])
        self.dlc_proc_dir.set("")
        self.dlc_proc_name_entry["values"] = tuple()
        self.dlc_proc_name.set("")
        self.dlc_options_entry["values"] = tuple(self.cfg["dlc_options"].keys()) + (
            "Add DLC",
        )
        self.dlc_option.set("")
        self.subject_entry["values"] = tuple(self.cfg["subjects"])
        self.subject.set("")
        self.directory_entry["values"] = tuple(self.cfg["directories"])
        self.directory.set("")

    def remove_config(self):
        """ Remove configuration
        """

        cfg_name = self.cfg_name.get()
        delete_setup = messagebox.askyesnocancel(
            "Delete Config Permanently?",
            "Would you like to delete the configuration {} permanently (yes),\nremove the setup from the list for this session (no),\nor neither (cancel).".format(
                cfg_name
            ),
            parent=self.window,
        )
        if delete_setup is not None:
            if delete_setup:
                os.remove(self.get_config_path(cfg_name))
            self.cfg_list.remove(cfg_name)
            self.cfg_entry["values"] = tuple(self.cfg_list) + ("Create New Setup",)
            self.cfg_name.set("")

    def get_camera_names(self):
        """ Get camera names from configuration as a tuple
        """

        return tuple(self.cfg["cameras"].keys())

    def init_cam(self):
        """ Initialize camera
        """

        if self.cam_pose_proc is not None:
            messagebox.showerror(
                "Camera Exists",
                "Camera already exists! Please close current camera before initializing a new one.",
            )
            return

        this_cam = self.get_current_camera()

        if not this_cam:

            messagebox.showerror(
                "No Camera",
                "No camera selected. Please select a camera before initializing.",
                parent=self.window,
            )

        else:

            if this_cam["type"] == "Add Camera":

                self.add_camera_window()
                return

            else:

                self.cam_setup_window = Toplevel(self.window)
                self.cam_setup_window.title("Setting up camera...")
                Label(
                    self.cam_setup_window, text="Setting up camera, please wait..."
                ).pack()
                self.cam_setup_window.update()

                cam_obj = getattr(camera, this_cam["type"])
                cam = cam_obj(**this_cam["params"])
                self.cam_pose_proc = CameraPoseProcess(cam)
                ret = self.cam_pose_proc.start_capture_process()

                if cam.use_tk_display:
                    self.set_display_window()

                self.cam_setup_window.destroy()

    def get_current_camera(self):
        """ Get dictionary of the current camera
        """

        if self.camera_name.get():
            if self.camera_name.get() == "Add Camera":
                return {"type": "Add Camera"}
            else:
                return self.cfg["cameras"][self.camera_name.get()]

    def set_camera_param(self, key, value):
        """ Set a camera parameter
        """

        self.cfg["cameras"][self.camera_name.get()]["params"][key] = value

    def add_camera_window(self):
        """ Create gui to add a camera
        """

        add_cam = Tk()
        cur_row = 0

        Label(add_cam, text="Type: ").grid(sticky="w", row=cur_row, column=0)
        self.cam_type = StringVar(add_cam)

        cam_types = [c[0] for c in inspect.getmembers(camera, inspect.isclass)]
        cam_types = [c for c in cam_types if (c != "Camera") & ("Error" not in c)]

        type_entry = Combobox(add_cam, textvariable=self.cam_type, state="readonly")
        type_entry["values"] = tuple(cam_types)
        type_entry.current(0)
        type_entry.grid(sticky="nsew", row=cur_row, column=1)
        cur_row += 1

        Label(add_cam, text="Name: ").grid(sticky="w", row=cur_row, column=0)
        self.new_cam_name = StringVar(add_cam)
        Entry(add_cam, textvariable=self.new_cam_name).grid(
            sticky="nsew", row=cur_row, column=1
        )
        cur_row += 1

        Button(
            add_cam, text="Add Camera", command=lambda: self.add_cam_to_list(add_cam)
        ).grid(sticky="nsew", row=cur_row, column=1)
        cur_row += 1

        Button(add_cam, text="Cancel", command=add_cam.destroy).grid(
            sticky="nsew", row=cur_row, column=1
        )

        add_cam.mainloop()

    def add_cam_to_list(self, gui):
        """ Add new camera to the camera list
        """

        self.cfg["cameras"][self.new_cam_name.get()] = {
            "type": self.cam_type.get(),
            "params": {},
        }
        self.camera_name.set(self.new_cam_name.get())
        self.camera_entry["values"] = self.get_camera_names() + ("Add Camera",)
        self.save_config()
        # messagebox.showinfo("Camera Added", "Camera has been added to the dropdown menu. Please edit camera settings before initializing the new camera.", parent=gui)
        gui.destroy()

    def edit_cam_settings(self):
        """ GUI window to edit camera settings
        """

        arg_names, arg_vals, arg_dtypes, arg_restrict = self.get_cam_args()

        settings_window = Toplevel(self.window)
        settings_window.title("Camera Settings")
        cur_row = 0
        combobox_width = 15

        entry_vars = []
        for n, v in zip(arg_names, arg_vals):

            Label(settings_window, text=n + ": ").grid(row=cur_row, column=0)

            if type(v) is list:
                v = [str(x) if x is not None else "" for x in v]
                v = ", ".join(v)
            else:
                v = v if v is not None else ""
            entry_vars.append(StringVar(settings_window, value=str(v)))

            if n in arg_restrict.keys():
                restrict_vals = arg_restrict[n]
                if type(restrict_vals[0]) is list:
                    restrict_vals = [
                        ", ".join([str(i) for i in rv]) for rv in restrict_vals
                    ]
                Combobox(
                    settings_window,
                    textvariable=entry_vars[-1],
                    values=restrict_vals,
                    state="readonly",
                    width=combobox_width,
                ).grid(sticky="nsew", row=cur_row, column=1)
            else:
                Entry(settings_window, textvariable=entry_vars[-1]).grid(
                    sticky="nsew", row=cur_row, column=1
                )

            cur_row += 1

        cur_row += 1
        Button(
            settings_window,
            text="Update",
            command=lambda: self.update_camera_settings(
                arg_names, entry_vars, arg_dtypes, settings_window
            ),
        ).grid(sticky="nsew", row=cur_row, column=1)
        cur_row += 1
        Button(settings_window, text="Cancel", command=settings_window.destroy).grid(
            sticky="nsew", row=cur_row, column=1
        )

        _, row_count = settings_window.grid_size()
        for r in range(row_count):
            settings_window.grid_rowconfigure(r, minsize=20)

        settings_window.mainloop()

    def get_cam_args(self):
        """ Get arguments for the new camera
        """

        this_cam = self.get_current_camera()
        cam_obj = getattr(camera, this_cam["type"])
        arg_restrict = cam_obj.arg_restrictions()

        cam_args = inspect.getfullargspec(cam_obj)
        n_args = len(cam_args[0][1:])
        n_vals = len(cam_args[3])
        arg_names = []
        arg_vals = []
        arg_dtype = []
        for i in range(n_args):
            arg_names.append(cam_args[0][i + 1])

            if arg_names[i] in this_cam["params"].keys():
                val = this_cam["params"][arg_names[i]]
            else:
                val = None if i < n_args - n_vals else cam_args[3][n_vals - n_args + i]
            arg_vals.append(val)

            dt_val = val if i < n_args - n_vals else cam_args[3][n_vals - n_args + i]
            dt = type(dt_val) if type(dt_val) is not list else type(dt_val[0])
            arg_dtype.append(dt)

        return arg_names, arg_vals, arg_dtype, arg_restrict

    def update_camera_settings(self, names, entries, dtypes, gui):
        """ Update camera settings from values input in settings GUI
        """

        gui.destroy()

        for name, entry, dt in zip(names, entries, dtypes):
            val = entry.get()
            val = val.split(",")
            val = [v.strip() for v in val]
            try:
                if dt is bool:
                    val = [True if v == "True" else False for v in val]
                else:
                    val = [dt(v) if v else None for v in val]
            except TypeError:
                pass
            val = val if len(val) > 1 else val[0]
            self.set_camera_param(name, val)

        self.save_config()

    def set_display_window(self):
        """ Create a video display window
        """

        self.display_window = Toplevel(self.window)
        self.display_frame_label = Label(self.display_window)
        self.display_frame_label.pack()
        self.display_frame()

    def set_display_colors(self, bodyparts):
        """ Set colors for keypoints

        Parameters
        ----------
        bodyparts : int
            the number of keypoints
        """

        all_colors = getattr(cc, self.display_cmap)
        self.display_colors = all_colors[:: int(len(all_colors) / bodyparts)]

    def display_frame(self):
        """ Display a frame in display window
        """

        if self.cam_pose_proc and self.display_window:

            frame = self.cam_pose_proc.get_display_frame()

            if frame is not None:

                img = Image.fromarray(frame)
                if frame.ndim == 3:
                    b, g, r = img.split()
                    img = Image.merge("RGB", (r, g, b))

                pose = (
                    self.cam_pose_proc.get_display_pose()
                    if self.display_keypoints.get()
                    else None
                )

                if pose is not None:

                    im_size = (frame.shape[1], frame.shape[0])

                    if not self.display_colors:
                        self.set_display_colors(pose.shape[0])

                    img_draw = ImageDraw.Draw(img)

                    for i in range(pose.shape[0]):
                        if pose[i, 2] > self.display_lik_thresh:
                            try:
                                x0 = (
                                    pose[i, 0] - self.display_radius
                                    if pose[i, 0] - self.display_radius > 0
                                    else 0
                                )
                                x1 = (
                                    pose[i, 0] + self.display_radius
                                    if pose[i, 0] + self.display_radius < im_size[1]
                                    else im_size[1]
                                )
                                y0 = (
                                    pose[i, 1] - self.display_radius
                                    if pose[i, 1] - self.display_radius > 0
                                    else 0
                                )
                                y1 = (
                                    pose[i, 1] + self.display_radius
                                    if pose[i, 1] + self.display_radius < im_size[0]
                                    else im_size[0]
                                )
                                coords = [x0, y0, x1, y1]
                                img_draw.ellipse(
                                    coords,
                                    fill=self.display_colors[i],
                                    outline=self.display_colors[i],
                                )
                            except Exception as e:
                                print(e)

                imgtk = ImageTk.PhotoImage(image=img)
                self.display_frame_label.imgtk = imgtk
                self.display_frame_label.configure(image=imgtk)

            self.display_frame_label.after(10, self.display_frame)

    def change_display_keypoints(self):
        """ Toggle display keypoints. If turning on, set display options. If turning off, destroy display window
        """

        if self.display_keypoints.get():

            display_options = self.cfg["dlc_display_options"][
                self.dlc_option.get()
            ].copy()
            self.display_cmap = display_options["cmap"]
            self.display_radius = display_options["radius"]
            self.display_lik_thresh = display_options["lik_thresh"]

            if not self.display_window:
                self.set_display_window()

        else:

            if self.cam_pose_proc is not None:
                if not self.cam_pose_proc.device.use_tk_display:
                    if self.display_window:
                        self.display_window.destroy()
                        self.display_window = None
                        self.display_colors = None

    def edit_dlc_display(self):

        display_options = self.cfg["dlc_display_options"][self.dlc_option.get()]

        dlc_display_settings = {
            "color map": {
                "value": display_options["cmap"],
                "dtype": str,
                "restriction": ["bgy", "kbc", "bmw", "bmy", "kgy", "fire"],
            },
            "radius": {"value": display_options["radius"], "dtype": int},
            "likelihood threshold": {
                "value": display_options["lik_thresh"],
                "dtype": float,
            },
        }

        dlc_display_gui = SettingsWindow(
            title="Edit DLC Display Settings",
            settings=dlc_display_settings,
            parent=self.window,
        )

        dlc_display_gui.mainloop()
        display_settings = dlc_display_gui.get_values()

        display_options["cmap"] = display_settings["color map"]
        display_options["radius"] = display_settings["radius"]
        display_options["lik_thresh"] = display_settings["likelihood threshold"]

        self.display_cmap = display_options["cmap"]
        self.display_radius = display_options["radius"]
        self.display_lik_thresh = display_options["lik_thresh"]

        self.cfg["dlc_display_options"][self.dlc_option.get()] = display_options
        self.save_config()

    def close_camera(self):
        """ Close capture process and display
        """

        if self.cam_pose_proc:
            if self.display_window is not None:
                self.display_window.destroy()
                self.display_window = None
            ret = self.cam_pose_proc.stop_capture_process()

        self.cam_pose_proc = None

    def change_dlc_option(self, event=None):

        if self.dlc_option.get() == "Add DLC":
            self.edit_dlc_settings(True)

    def edit_dlc_settings(self, new=False):

        if new:
            cur_set = self.empty_dlc_settings()
        else:
            cur_set = self.cfg["dlc_options"][self.dlc_option.get()].copy()
            cur_set["name"] = self.dlc_option.get()
            cur_set["cropping"] = (
                ", ".join([str(c) for c in cur_set["cropping"]])
                if cur_set["cropping"]
                else ""
            )
            cur_set["dynamic"] = ", ".join([str(d) for d in cur_set["dynamic"]])
            cur_set["mode"] = (
                "Optimize Latency" if "mode" not in cur_set else cur_set["mode"]
            )

        self.dlc_settings_window = Toplevel(self.window)
        self.dlc_settings_window.title("DLC Settings")
        cur_row = 0

        Label(self.dlc_settings_window, text="Name: ").grid(
            sticky="w", row=cur_row, column=0
        )
        self.dlc_settings_name = StringVar(
            self.dlc_settings_window, value=cur_set["name"]
        )
        Entry(self.dlc_settings_window, textvariable=self.dlc_settings_name).grid(
            sticky="nsew", row=cur_row, column=1
        )
        cur_row += 1

        Label(self.dlc_settings_window, text="Model Path: ").grid(
            sticky="w", row=cur_row, column=0
        )
        self.dlc_settings_model_path = StringVar(
            self.dlc_settings_window, value=cur_set["model_path"]
        )
        Entry(self.dlc_settings_window, textvariable=self.dlc_settings_model_path).grid(
            sticky="nsew", row=cur_row, column=1
        )
        Button(
            self.dlc_settings_window, text="Browse", command=self.browse_dlc_path
        ).grid(sticky="nsew", row=cur_row, column=2)
        cur_row += 1

        Label(self.dlc_settings_window, text="Model Type: ").grid(
            sticky="w", row=cur_row, column=0
        )
        self.dlc_settings_model_type = StringVar(
            self.dlc_settings_window, value=cur_set["model_type"]
        )
        Combobox(
            self.dlc_settings_window,
            textvariable=self.dlc_settings_model_type,
            value=["base", "tensorrt", "tflite"],
            state="readonly",
        ).grid(sticky="nsew", row=cur_row, column=1)
        cur_row += 1

        Label(self.dlc_settings_window, text="Precision: ").grid(
            sticky="w", row=cur_row, column=0
        )
        self.dlc_settings_precision = StringVar(
            self.dlc_settings_window, value=cur_set["precision"]
        )
        Combobox(
            self.dlc_settings_window,
            textvariable=self.dlc_settings_precision,
            value=["FP32", "FP16", "INT8"],
            state="readonly",
        ).grid(sticky="nsew", row=cur_row, column=1)
        cur_row += 1

        Label(self.dlc_settings_window, text="Cropping: ").grid(
            sticky="w", row=cur_row, column=0
        )
        self.dlc_settings_cropping = StringVar(
            self.dlc_settings_window, value=cur_set["cropping"]
        )
        Entry(self.dlc_settings_window, textvariable=self.dlc_settings_cropping).grid(
            sticky="nsew", row=cur_row, column=1
        )
        cur_row += 1

        Label(self.dlc_settings_window, text="Dynamic: ").grid(
            sticky="w", row=cur_row, column=0
        )
        self.dlc_settings_dynamic = StringVar(
            self.dlc_settings_window, value=cur_set["dynamic"]
        )
        Entry(self.dlc_settings_window, textvariable=self.dlc_settings_dynamic).grid(
            sticky="nsew", row=cur_row, column=1
        )
        cur_row += 1

        Label(self.dlc_settings_window, text="Resize: ").grid(
            sticky="w", row=cur_row, column=0
        )
        self.dlc_settings_resize = StringVar(
            self.dlc_settings_window, value=cur_set["resize"]
        )
        Entry(self.dlc_settings_window, textvariable=self.dlc_settings_resize).grid(
            sticky="nsew", row=cur_row, column=1
        )
        cur_row += 1

        Label(self.dlc_settings_window, text="Mode: ").grid(
            sticky="w", row=cur_row, column=0
        )
        self.dlc_settings_mode = StringVar(
            self.dlc_settings_window, value=cur_set["mode"]
        )
        Combobox(
            self.dlc_settings_window,
            textvariable=self.dlc_settings_mode,
            state="readonly",
            values=["Optimize Latency", "Optimize Rate"],
        ).grid(sticky="nsew", row=cur_row, column=1)
        cur_row += 1

        Button(
            self.dlc_settings_window, text="Update", command=self.update_dlc_settings
        ).grid(sticky="nsew", row=cur_row, column=1)
        Button(
            self.dlc_settings_window,
            text="Cancel",
            command=self.dlc_settings_window.destroy,
        ).grid(sticky="nsew", row=cur_row, column=2)

    def empty_dlc_settings(self):

        return {
            "name": "",
            "model_path": "",
            "model_type": "base",
            "precision": "FP32",
            "cropping": "",
            "dynamic": "False, 0.5, 10",
            "resize": "1.0",
            "mode": "Optimize Latency",
        }

    def browse_dlc_path(self):
        """ Open file browser to select DLC exported model directory
        """

        new_dlc_path = filedialog.askdirectory(parent=self.dlc_settings_window)
        if new_dlc_path:
            self.dlc_settings_model_path.set(new_dlc_path)

    def update_dlc_settings(self):
        """ Update DLC settings for the current dlc option from DLC Settings GUI
        """

        precision = (
            self.dlc_settings_precision.get()
            if self.dlc_settings_precision.get()
            else "FP32"
        )

        crop_warn = False
        dlc_crop = self.dlc_settings_cropping.get()
        if dlc_crop:
            try:
                dlc_crop = dlc_crop.split(",")
                assert len(dlc_crop) == 4
                dlc_crop = [int(c) for c in dlc_crop]
            except Exception:
                crop_warn = True
                dlc_crop = None
        else:
            dlc_crop = None

        try:
            dlc_dynamic = self.dlc_settings_dynamic.get().replace(" ", "")
            dlc_dynamic = dlc_dynamic.split(",")
            dlc_dynamic[0] = True if dlc_dynamic[0] == "True" else False
            dlc_dynamic[1] = float(dlc_dynamic[1])
            dlc_dynamic[2] = int(dlc_dynamic[2])
            dlc_dynamic = tuple(dlc_dynamic)
            dyn_warn = False
        except Exception:
            dyn_warn = True
            dlc_dynamic = (False, 0.5, 10)

        dlc_resize = (
            float(self.dlc_settings_resize.get())
            if self.dlc_settings_resize.get()
            else None
        )
        dlc_mode = self.dlc_settings_mode.get()

        warn_msg = ""
        if crop_warn:
            warn_msg += "DLC Cropping was not set properly. Using default cropping parameters...\n"
        if dyn_warn:
            warn_msg += "DLC Dynamic Cropping was not set properly. Using default dynamic cropping parameters..."
        if warn_msg:
            messagebox.showerror(
                "DLC Settings Error", warn_msg, parent=self.dlc_settings_window
            )

        self.cfg["dlc_options"][self.dlc_settings_name.get()] = {
            "model_path": self.dlc_settings_model_path.get(),
            "model_type": self.dlc_settings_model_type.get(),
            "precision": precision,
            "cropping": dlc_crop,
            "dynamic": dlc_dynamic,
            "resize": dlc_resize,
            "mode": dlc_mode,
        }

        if self.dlc_settings_name.get() not in self.cfg["dlc_display_options"]:
            self.cfg["dlc_display_options"][self.dlc_settings_name.get()] = {
                "cmap": "bgy",
                "radius": 3,
                "lik_thresh": 0.5,
            }

        self.save_config()
        self.dlc_options_entry["values"] = tuple(self.cfg["dlc_options"].keys()) + (
            "Add DLC",
        )
        self.dlc_option.set(self.dlc_settings_name.get())
        self.dlc_settings_window.destroy()

    def remove_dlc_option(self):
        """ Delete DLC Option from config
        """

        del self.cfg["dlc_options"][self.dlc_option.get()]
        del self.cfg["dlc_display_options"][self.dlc_option.get()]
        self.dlc_options_entry["values"] = tuple(self.cfg["dlc_options"].keys()) + (
            "Add DLC",
        )
        self.dlc_option.set("")
        self.save_config()

    def init_dlc(self):
        """ Initialize DLC Live object
        """

        self.stop_pose()

        self.dlc_setup_window = Toplevel(self.window)
        self.dlc_setup_window.title("Setting up DLC...")
        Label(self.dlc_setup_window, text="Setting up DLC, please wait...").pack()
        self.dlc_setup_window.after(10, self.start_pose)
        self.dlc_setup_window.mainloop()

    def start_pose(self):

        dlc_params = self.cfg["dlc_options"][self.dlc_option.get()].copy()
        dlc_params["processor"] = self.dlc_proc_params
        ret = self.cam_pose_proc.start_pose_process(dlc_params)
        self.dlc_setup_window.destroy()

    def stop_pose(self):
        """ Stop pose process
        """

        if self.cam_pose_proc:
            ret = self.cam_pose_proc.stop_pose_process()

    def add_subject(self):
        new_sub = self.subject.get()
        if new_sub:
            if new_sub not in self.cfg["subjects"]:
                self.cfg["subjects"].append(new_sub)
                self.subject_entry["values"] = tuple(self.cfg["subjects"])
                self.save_config()

    def remove_subject(self):

        self.cfg["subjects"].remove(self.subject.get())
        self.subject_entry["values"] = self.cfg["subjects"]
        self.save_config()
        self.subject.set("")

    def browse_directory(self):

        new_dir = filedialog.askdirectory(parent=self.window)
        if new_dir:
            self.directory.set(new_dir)
            ask_add_dir = Tk()
            Label(
                ask_add_dir,
                text="Would you like to add this directory to dropdown list?",
            ).pack()
            Button(
                ask_add_dir, text="Yes", command=lambda: self.add_directory(ask_add_dir)
            ).pack()
            Button(ask_add_dir, text="No", command=ask_add_dir.destroy).pack()

    def add_directory(self, window):

        window.destroy()
        if self.directory.get() not in self.cfg["directories"]:
            self.cfg["directories"].append(self.directory.get())
            self.directory_entry["values"] = self.cfg["directories"]
            self.save_config()

    def save_config(self, notify=False):

        json.dump(self.cfg, open(self.cfg_file, "w"))
        if notify:
            messagebox.showinfo(
                title="Config file saved",
                message="Configuration file has been saved...",
                parent=self.window,
            )

    def remove_cam_cfg(self):

        if self.camera_name.get() != "Add Camera":
            delete = messagebox.askyesno(
                title="Delete Camera?",
                message="Are you sure you want to delete '%s'?"
                % self.camera_name.get(),
                parent=self.window,
            )
            if delete:
                del self.cfg["cameras"][self.camera_name.get()]
                self.camera_entry["values"] = self.get_camera_names() + ("Add Camera",)
                self.camera_name.set("")
                self.save_config()

    def browse_dlc_processor(self):

        new_dir = filedialog.askdirectory(parent=self.window)
        if new_dir:
            self.dlc_proc_dir.set(new_dir)
            self.update_dlc_proc_list()

            if new_dir not in self.cfg["processor_dir"]:
                if messagebox.askyesno(
                    "Add to dropdown",
                    "Would you like to add this directory to dropdown list?",
                ):
                    self.cfg["processor_dir"].append(new_dir)
                    self.dlc_proc_dir_entry["values"] = tuple(self.cfg["processor_dir"])
                    self.save_config()

    def rem_dlc_proc_dir(self):

        if self.dlc_proc_dir.get() in self.cfg["processor_dir"]:
            self.cfg["processor_dir"].remove(self.dlc_proc_dir.get())
            self.save_config()
        self.dlc_proc_dir_entry["values"] = tuple(self.cfg["processor_dir"])
        self.dlc_proc_dir.set("")

    def update_dlc_proc_list(self, event=None):

        ### if dlc proc module already initialized, delete module and remove from path ###

        self.processor_list = []

        if self.dlc_proc_dir.get():

            if hasattr(self, "dlc_proc_module"):
                sys.path.remove(sys.path[0])

            new_path = os.path.normpath(os.path.dirname(self.dlc_proc_dir.get()))
            if new_path not in sys.path:
                sys.path.insert(0, new_path)

            new_mod = os.path.basename(self.dlc_proc_dir.get())
            if new_mod in sys.modules:
                del sys.modules[new_mod]

            ### load new module ###

            processor_spec = importlib.util.find_spec(
                os.path.basename(self.dlc_proc_dir.get())
            )
            try:
                self.dlc_proc_module = importlib.util.module_from_spec(processor_spec)
                processor_spec.loader.exec_module(self.dlc_proc_module)
                # self.processor_list = inspect.getmembers(self.dlc_proc_module, inspect.isclass)
                self.processor_list = [
                    proc for proc in dir(self.dlc_proc_module) if "__" not in proc
                ]
            except AttributeError:
                if hasattr(self, "window"):
                    messagebox.showerror(
                        "Failed to load processors!",
                        "Failed to load processors from directory = "
                        + self.dlc_proc_dir.get()
                        + ".\nPlease select a different directory.",
                        parent=self.window,
                    )

            self.dlc_proc_name_entry["values"] = tuple(self.processor_list)

    def set_proc(self):

        # proc_param_dict = {}
        # for i in range(1, len(self.proc_param_names)):
        #     proc_param_dict[self.proc_param_names[i]] = self.proc_param_default_types[i](self.proc_param_values[i-1].get())

        # if self.dlc_proc_dir.get() not in self.cfg['processor_args']:
        #     self.cfg['processor_args'][self.dlc_proc_dir.get()] = {}
        # self.cfg['processor_args'][self.dlc_proc_dir.get()][self.dlc_proc_name.get()] = proc_param_dict
        # self.save_config()

        # self.dlc_proc = self.proc_object(**proc_param_dict)
        proc_object = getattr(self.dlc_proc_module, self.dlc_proc_name.get())
        self.dlc_proc_params = {"object": proc_object}
        self.dlc_proc_params.update(
            self.cfg["processor_args"][self.dlc_proc_dir.get()][
                self.dlc_proc_name.get()
            ]
        )

    def clear_proc(self):

        self.dlc_proc_params = None

    def edit_proc(self):

        ### get default args: load module and read arguments ###

        self.proc_object = getattr(self.dlc_proc_module, self.dlc_proc_name.get())
        def_args = inspect.getargspec(self.proc_object)
        self.proc_param_names = def_args[0]
        self.proc_param_default_values = def_args[3]
        self.proc_param_default_types = [
            type(v) if type(v) is not list else [type(v[0])] for v in def_args[3]
        ]
        for i in range(len(def_args[0]) - len(def_args[3])):
            self.proc_param_default_values = ("",) + self.proc_param_default_values
            self.proc_param_default_types = [str] + self.proc_param_default_types

        ### check for existing settings in config ###

        old_args = {}
        if self.dlc_proc_dir.get() in self.cfg["processor_args"]:
            if (
                self.dlc_proc_name.get()
                in self.cfg["processor_args"][self.dlc_proc_dir.get()]
            ):
                old_args = self.cfg["processor_args"][self.dlc_proc_dir.get()][
                    self.dlc_proc_name.get()
                ].copy()
        else:
            self.cfg["processor_args"][self.dlc_proc_dir.get()] = {}

        ### get dictionary of arguments ###

        proc_args_dict = {}
        for i in range(1, len(self.proc_param_names)):

            if self.proc_param_names[i] in old_args:
                this_value = old_args[self.proc_param_names[i]]
            else:
                this_value = self.proc_param_default_values[i]

            proc_args_dict[self.proc_param_names[i]] = {
                "value": this_value,
                "dtype": self.proc_param_default_types[i],
            }

        proc_args_gui = SettingsWindow(
            title="DLC Processor Settings", settings=proc_args_dict, parent=self.window
        )
        proc_args_gui.mainloop()

        self.cfg["processor_args"][self.dlc_proc_dir.get()][
            self.dlc_proc_name.get()
        ] = proc_args_gui.get_values()
        self.save_config()

    def init_session(self):

        ### check if video is currently open ###

        if self.record_on.get() > -1:
            messagebox.showerror(
                "Session Open",
                "Session is currently open! Please release the current video (click 'Save Video' of 'Delete Video', even if no frames have been recorded) before setting up a new one.",
                parent=self.window,
            )
            return

        ### check if camera is already set up ###

        if not self.cam_pose_proc:
            messagebox.showerror(
                "No Camera",
                "No camera is found! Please initialize a camera before setting up the video.",
                parent=self.window,
            )
            return

        ### set up session window

        self.session_setup_window = Toplevel(self.window)
        self.session_setup_window.title("Setting up session...")
        Label(
            self.session_setup_window, text="Setting up session, please wait..."
        ).pack()
        self.session_setup_window.after(10, self.start_writer)
        self.session_setup_window.mainloop()

    def start_writer(self):

        ### set up file name (get date and create directory)

        dt = datetime.datetime.now()
        date = f"{dt.year:04d}-{dt.month:02d}-{dt.day:02d}"
        self.out_dir = self.directory.get()
        if not os.path.isdir(os.path.normpath(self.out_dir)):
            os.makedirs(os.path.normpath(self.out_dir))

        ### create output file names

        self.base_name = os.path.normpath(
            f"{self.out_dir}/{self.camera_name.get().replace(' ', '')}_{self.subject.get()}_{date}_{self.attempt.get()}"
        )
        # self.vid_file = os.path.normpath(self.out_dir + '/VIDEO_' + self.base_name + '.avi')
        # self.ts_file = os.path.normpath(self.out_dir + '/TIMESTAMPS_' + self.base_name + '.pickle')
        # self.dlc_file = os.path.normpath(self.out_dir + '/DLC_' + self.base_name + '.h5')
        # self.proc_file = os.path.normpath(self.out_dir + '/PROC_' + self.base_name + '.pickle')

        ### check if files already exist

        fs = glob.glob(f"{self.base_name}*")
        if len(fs) > 0:
            overwrite = messagebox.askyesno(
                "Files Exist",
                "Files already exist with attempt number = {}. Would you like to overwrite the file?".format(
                    self.attempt.get()
                ),
                parent=self.session_setup_window,
            )
            if not overwrite:
                return

        ### start writer

        ret = self.cam_pose_proc.start_writer_process(self.base_name)

        self.session_setup_window.destroy()

        ### set GUI to Ready

        self.record_on.set(0)

    def start_record(self):
        """ Issues command to start recording frames and poses
        """

        ret = False
        if self.cam_pose_proc is not None:
            ret = self.cam_pose_proc.start_record()

        if not ret:
            messagebox.showerror(
                "Recording Not Ready",
                "Recording has not been set up. Please make sure a camera and session have been initialized.",
                parent=self.window,
            )
            self.record_on.set(-1)

    def stop_record(self):
        """ Issues command to stop recording frames and poses
        """

        if self.cam_pose_proc is not None:
            ret = self.cam_pose_proc.stop_record()
            self.record_on.set(0)

    def save_vid(self, delete=False):
        """ Saves video, timestamp, and DLC files

        Parameters
        ----------
        delete : bool, optional
            flag to delete created files, by default False
        """

        ### perform checks ###

        if self.cam_pose_proc is None:
            messagebox.showwarning(
                "No Camera",
                "Camera has not yet been initialized, no video recorded.",
                parent=self.window,
            )
            return

        elif self.record_on.get() == -1:
            messagebox.showwarning(
                "No Video File",
                "Video was not set up, no video recorded.",
                parent=self.window,
            )
            return

        elif self.record_on.get() == 1:
            messagebox.showwarning(
                "Active Recording",
                "You are currently recording a video. Please stop the video before saving.",
                parent=self.window,
            )
            return

        elif delete:
            delete = messagebox.askokcancel(
                "Delete Video?", "Do you wish to delete the video?", parent=self.window
            )

        ### save or delete video ###

        if delete:
            ret = self.cam_pose_proc.stop_writer_process(save=False)
            messagebox.showinfo(
                "Video Deleted",
                "Video and timestamp files have been deleted.",
                parent=self.window,
            )
        else:
            ret = self.cam_pose_proc.stop_writer_process(save=True)
            ret_pose = self.cam_pose_proc.save_pose(self.base_name)
            if ret:
                if ret_pose:
                    messagebox.showinfo(
                        "Files Saved",
                        "Video, timestamp, and DLC Files have been saved.",
                    )
                else:
                    messagebox.showinfo(
                        "Files Saved", "Video and timestamp files have been saved."
                    )
            else:
                messagebox.showwarning(
                    "No Frames Recorded",
                    "No frames were recorded, video was deleted",
                    parent=self.window,
                )

        self.record_on.set(-1)

    def closeGUI(self):

        if self.cam_pose_proc:
            ret = self.cam_pose_proc.stop_writer_process()
            ret = self.cam_pose_proc.stop_pose_process()
            ret = self.cam_pose_proc.stop_capture_process()

        self.window.destroy()

    def createGUI(self):

        ### initialize window ###

        self.window = Tk()
        self.window.title("DeepLabCut Live")
        cur_row = 0
        combobox_width = 15

        ### select cfg file
        if len(self.cfg_list) > 0:
            initial_cfg = self.cfg_list[0]
        else:
            initial_cfg = ""

        Label(self.window, text="Config: ").grid(sticky="w", row=cur_row, column=0)
        self.cfg_name = StringVar(self.window, value=initial_cfg)
        self.cfg_entry = Combobox(
            self.window, textvariable=self.cfg_name, width=combobox_width
        )
        self.cfg_entry["values"] = tuple(self.cfg_list) + ("Create New Config",)
        self.cfg_entry.bind("<<ComboboxSelected>>", self.change_config)
        self.cfg_entry.grid(sticky="nsew", row=cur_row, column=1)
        Button(self.window, text="Remove Config", command=self.remove_config).grid(
            sticky="nsew", row=cur_row, column=2
        )

        self.get_config(initial_cfg)

        cur_row += 2

        ### select camera ###

        # camera entry
        Label(self.window, text="Camera: ").grid(sticky="w", row=cur_row, column=0)
        self.camera_name = StringVar(self.window)
        self.camera_entry = Combobox(self.window, textvariable=self.camera_name)
        cam_names = self.get_camera_names()
        self.camera_entry["values"] = cam_names + ("Add Camera",)
        if cam_names:
            self.camera_entry.current(0)
        self.camera_entry.grid(sticky="nsew", row=cur_row, column=1)
        Button(self.window, text="Init Cam", command=self.init_cam).grid(
            sticky="nsew", row=cur_row, column=2
        )
        cur_row += 1

        Button(
            self.window, text="Edit Camera Settings", command=self.edit_cam_settings
        ).grid(sticky="nsew", row=cur_row, column=1)
        Button(self.window, text="Close Camera", command=self.close_camera).grid(
            sticky="nsew", row=cur_row, column=2
        )
        cur_row += 1
        Button(self.window, text="Remove Camera", command=self.remove_cam_cfg).grid(
            sticky="nsew", row=cur_row, column=2
        )

        cur_row += 2

        ### set up proc ###

        Label(self.window, text="Processor Dir: ").grid(
            sticky="w", row=cur_row, column=0
        )
        self.dlc_proc_dir = StringVar(self.window)
        self.dlc_proc_dir_entry = Combobox(self.window, textvariable=self.dlc_proc_dir)
        self.dlc_proc_dir_entry["values"] = tuple(self.cfg["processor_dir"])
        if len(self.cfg["processor_dir"]) > 0:
            self.dlc_proc_dir_entry.current(0)
        self.dlc_proc_dir_entry.bind("<<ComboboxSelected>>", self.update_dlc_proc_list)
        self.dlc_proc_dir_entry.grid(sticky="nsew", row=cur_row, column=1)
        Button(self.window, text="Browse", command=self.browse_dlc_processor).grid(
            sticky="nsew", row=cur_row, column=2
        )
        Button(self.window, text="Remove Proc Dir", command=self.rem_dlc_proc_dir).grid(
            sticky="nsew", row=cur_row + 1, column=2
        )
        cur_row += 2

        Label(self.window, text="Processor: ").grid(sticky="w", row=cur_row, column=0)
        self.dlc_proc_name = StringVar(self.window)
        self.dlc_proc_name_entry = Combobox(
            self.window, textvariable=self.dlc_proc_name
        )
        self.update_dlc_proc_list()
        # self.dlc_proc_name_entry['values'] = tuple(self.processor_list) # tuple([c[0] for c in inspect.getmembers(processor, inspect.isclass)])
        if len(self.processor_list) > 0:
            self.dlc_proc_name_entry.current(0)
        self.dlc_proc_name_entry.grid(sticky="nsew", row=cur_row, column=1)
        Button(self.window, text="Set Proc", command=self.set_proc).grid(
            sticky="nsew", row=cur_row, column=2
        )
        Button(self.window, text="Edit Proc Settings", command=self.edit_proc).grid(
            sticky="nsew", row=cur_row + 1, column=1
        )
        Button(self.window, text="Clear Proc", command=self.clear_proc).grid(
            sticky="nsew", row=cur_row + 1, column=2
        )

        cur_row += 3

        ### set up dlc live ###

        Label(self.window, text="DeepLabCut: ").grid(sticky="w", row=cur_row, column=0)
        self.dlc_option = StringVar(self.window)
        self.dlc_options_entry = Combobox(self.window, textvariable=self.dlc_option)
        self.dlc_options_entry["values"] = tuple(self.cfg["dlc_options"].keys()) + (
            "Add DLC",
        )
        self.dlc_options_entry.bind("<<ComboboxSelected>>", self.change_dlc_option)
        self.dlc_options_entry.grid(sticky="nsew", row=cur_row, column=1)
        Button(self.window, text="Init DLC", command=self.init_dlc).grid(
            sticky="nsew", row=cur_row, column=2
        )
        cur_row += 1

        Button(
            self.window, text="Edit DLC Settings", command=self.edit_dlc_settings
        ).grid(sticky="nsew", row=cur_row, column=1)
        Button(self.window, text="Stop DLC", command=self.stop_pose).grid(
            sticky="nsew", row=cur_row, column=2
        )
        cur_row += 1

        self.display_keypoints = BooleanVar(self.window, value=False)
        Checkbutton(
            self.window,
            text="Display DLC Keypoints",
            variable=self.display_keypoints,
            indicatoron=0,
            command=self.change_display_keypoints,
        ).grid(sticky="nsew", row=cur_row, column=1)
        Button(self.window, text="Remove DLC", command=self.remove_dlc_option).grid(
            sticky="nsew", row=cur_row, column=2
        )
        cur_row += 1

        Button(
            self.window, text="Edit DLC Display Settings", command=self.edit_dlc_display
        ).grid(sticky="nsew", row=cur_row, column=1)

        cur_row += 2

        ### set up session ###

        # subject
        Label(self.window, text="Subject: ").grid(sticky="w", row=cur_row, column=0)
        self.subject = StringVar(self.window)
        self.subject_entry = Combobox(self.window, textvariable=self.subject)
        self.subject_entry["values"] = self.cfg["subjects"]
        self.subject_entry.grid(sticky="nsew", row=cur_row, column=1)
        Button(self.window, text="Add Subject", command=self.add_subject).grid(
            sticky="nsew", row=cur_row, column=2
        )
        cur_row += 1

        # attempt
        Label(self.window, text="Attempt: ").grid(sticky="w", row=cur_row, column=0)
        self.attempt = StringVar(self.window)
        self.attempt_entry = Combobox(self.window, textvariable=self.attempt)
        self.attempt_entry["values"] = tuple(range(1, 10))
        self.attempt_entry.current(0)
        self.attempt_entry.grid(sticky="nsew", row=cur_row, column=1)
        Button(self.window, text="Remove Subject", command=self.remove_subject).grid(
            sticky="nsew", row=cur_row, column=2
        )
        cur_row += 1

        # out directory
        Label(self.window, text="Directory: ").grid(sticky="w", row=cur_row, column=0)
        self.directory = StringVar(self.window)
        self.directory_entry = Combobox(self.window, textvariable=self.directory)
        if self.cfg["directories"]:
            self.directory_entry["values"] = self.cfg["directories"]
            self.directory_entry.current(0)
        self.directory_entry.grid(sticky="nsew", row=cur_row, column=1)
        Button(self.window, text="Browse", command=self.browse_directory).grid(
            sticky="nsew", row=cur_row, column=2
        )
        cur_row += 1

        # set up session
        Button(self.window, text="Set Up Session", command=self.init_session).grid(
            sticky="nsew", row=cur_row, column=1
        )
        cur_row += 2

        ### control recording ###

        Label(self.window, text="Record: ").grid(sticky="w", row=cur_row, column=0)
        self.record_on = IntVar(value=-1)
        Radiobutton(
            self.window,
            text="Ready",
            selectcolor="blue",
            indicatoron=0,
            variable=self.record_on,
            value=0,
            state="disabled",
        ).grid(stick="nsew", row=cur_row, column=1)
        Radiobutton(
            self.window,
            text="On",
            selectcolor="green",
            indicatoron=0,
            variable=self.record_on,
            value=1,
            command=self.start_record,
        ).grid(sticky="nsew", row=cur_row + 1, column=1)
        Radiobutton(
            self.window,
            text="Off",
            selectcolor="red",
            indicatoron=0,
            variable=self.record_on,
            value=-1,
            command=self.stop_record,
        ).grid(sticky="nsew", row=cur_row + 2, column=1)
        Button(self.window, text="Save Video", command=lambda: self.save_vid()).grid(
            sticky="nsew", row=cur_row + 1, column=2
        )
        Button(
            self.window, text="Delete Video", command=lambda: self.save_vid(delete=True)
        ).grid(sticky="nsew", row=cur_row + 2, column=2)

        cur_row += 4

        ### close program ###

        Button(self.window, text="Close", command=self.closeGUI).grid(
            sticky="nsew", row=cur_row, column=0, columnspan=2
        )

        ### configure size of empty rows

        _, row_count = self.window.grid_size()
        for r in range(row_count):
            self.window.grid_rowconfigure(r, minsize=20)

    def run(self):

        self.window.mainloop()


def main():

    # import multiprocess as mp
    # mp.set_start_method("spawn")

    dlc_live_gui = DLCLiveGUI()
    dlc_live_gui.run()



"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


import cv2
import time


class CameraError(Exception):
    """
    Exception for incorrect use of cameras
    """

    pass


class Camera(object):
    """ Base camera class. Controls image capture, writing images to video, pose estimation and image display. 
    
    Parameters
    ----------
    id : [type]
        camera id
    exposure : int, optional
        exposure time in microseconds, by default None
    gain : int, optional
        gain value, by default None
    rotate : [type], optional
        [description], by default None
    crop : list, optional
        camera cropping parameters: [left, right, top, bottom], by default None
    fps : float, optional
        frame rate in frames per second, by default None
    use_tk_display : bool, optional
        flag to use tk image display (if using GUI), by default False
    display_resize : float, optional
        factor to resize images if using opencv display (display is very slow for large images), by default None
    """

    @staticmethod
    def arg_restrictions():
        """ Returns a dictionary of arguments restrictions for DLCLiveGUI
        """

        return {}

    def __init__(
        self,
        id,
        resolution=None,
        exposure=None,
        gain=None,
        rotate=None,
        crop=None,
        fps=None,
        use_tk_display=False,
        display_resize=1.0,
    ):
        """ Constructor method
        """

        self.id = id
        self.exposure = exposure
        self.gain = gain
        self.rotate = rotate
        self.crop = [int(c) for c in crop] if crop else None
        self.set_im_size(resolution)
        self.fps = fps
        self.use_tk_display = use_tk_display
        self.display_resize = display_resize if display_resize else 1.0
        self.next_frame = 0

    def set_im_size(self, res):
        """[summary]
        
        Parameters
        ----------
        default : [, optional
            [description], by default None
        
        Raises
        ------
        DLCLiveCameraError
            throws error if resolution is not set
        """

        if not res:
            raise CameraError("Resolution is not set!")

        self.im_size = (
            (int(res[0]), int(res[1]))
            if self.crop is None
            else (self.crop[3] - self.crop[2], self.crop[1] - self.crop[0])
        )

    def set_capture_device(self):
        """ Sets frame capture device with desired properties
        """

        raise NotImplementedError

    def get_image_on_time(self):
        """ Gets an image from frame capture device at the appropriate time (according to fps).
        
        Returns
        -------
        `np.ndarray`
            image as a numpy array
        float
            timestamp at which frame was taken, obtained from :func:`time.time`
        """

        frame = None
        while frame is None:
            cur_time = time.time()
            if cur_time > self.next_frame:
                frame = self.get_image()
                timestamp = cur_time
                self.next_frame = max(
                    self.next_frame + 1.0 / self.fps, cur_time + 0.5 / self.fps
                )

        return frame, timestamp

    def get_image(self):
        """ Gets image from frame capture device
        """

        raise NotImplementedError

    def close_capture_device(self):
        """ Closes frame capture device
        """

        raise NotImplementedError



"""
DeepLabCut Toolbox (deeplabcut.org)
© A. & M. Mathis Labs

Licensed under GNU Lesser General Public License v3.0
"""


import multiprocess as mp
import threading
import time
import pandas as pd
import numpy as np

from dlclivegui import CameraProcess
from dlclivegui.queue import ClearableQueue, ClearableMPQueue


class DLCLiveProcessError(Exception):
    """
    Exception for incorrect use of DLC-live-GUI Process Manager
    """

    pass


class CameraPoseProcess(CameraProcess):
    """ Camera Process Manager class. Controls image capture, pose estimation and writing images to a video file in a background process.

    Parameters
    ----------
    device : :class:`cameracontrol.Camera`
        a camera object
    ctx : :class:`multiprocess.Context`
        multiprocessing context
    """

    def __init__(self, device, ctx=mp.get_context("spawn")):
        """ Constructor method
        """

        super().__init__(device, ctx)
        self.display_pose = None
        self.display_pose_queue = ClearableMPQueue(2, ctx=self.ctx)
        self.pose_process = None

    def start_pose_process(self, dlc_params, timeout=300):

        self.pose_process = self.ctx.Process(
            target=self._run_pose,
            args=(self.frame_shared, self.frame_time_shared, dlc_params),
            daemon=True,
        )
        self.pose_process.start()

        stime = time.time()
        while time.time() - stime < timeout:
            cmd = self.q_from_process.read()
            if cmd is not None:
                if (cmd[0] == "pose") and (cmd[1] == "start"):
                    return cmd[2]
                else:
                    self.q_to_process.write(cmd)

    def _run_pose(self, frame_shared, frame_time, dlc_params):

        res = self.device.im_size
        self.frame = np.frombuffer(frame_shared.get_obj(), dtype="uint8").reshape(
            res[1], res[0], 3
        )
        self.frame_time = np.frombuffer(frame_time.get_obj(), dtype="d")

        ret = self._open_dlc_live(dlc_params)
        self.q_from_process.write(("pose", "start", ret))

        self._pose_loop()
        self.q_from_process.write(("pose", "end"))

    def _open_dlc_live(self, dlc_params):

        from dlclive import DLCLive

        ret = False

        self.opt_rate = True if dlc_params.pop("mode") == "Optimize Rate" else False

        proc_params = dlc_params.pop("processor")
        if proc_params is not None:
            proc_obj = proc_params.pop("object", None)
            if proc_obj is not None:
                dlc_params["processor"] = proc_obj(**proc_params)

        self.dlc = DLCLive(**dlc_params)
        if self.frame is not None:
            self.dlc.init_inference(
                self.frame, frame_time=self.frame_time[0], record=False
            )
            self.poses = []
            self.pose_times = []
            self.pose_frame_times = []
            ret = True

        return ret

    def _pose_loop(self):
        """ Conduct pose estimation using deeplabcut-live in loop
        """

        run = True
        write = False
        frame_time = 0
        pose_time = 0
        end_time = time.time()

        while run:

            ref_time = frame_time if self.opt_rate else end_time

            if self.frame_time[0] > ref_time:

                frame = self.frame
                frame_time = self.frame_time[0]
                pose = self.dlc.get_pose(frame, frame_time=frame_time, record=write)
                pose_time = time.time()

                self.display_pose_queue.write(pose, clear=True)

                if write:
                    self.poses.append(pose)
                    self.pose_times.append(pose_time)
                    self.pose_frame_times.append(frame_time)

                cmd = self.q_to_process.read()
                if cmd is not None:
                    if cmd[0] == "pose":
                        if cmd[1] == "write":
                            write = cmd[2]
                            self.q_from_process.write(cmd)
                        elif cmd[1] == "save":
                            ret = self._save_pose(cmd[2])
                            self.q_from_process.write(cmd + (ret,))
                        elif cmd[1] == "end":
                            run = False
                    else:
                        self.q_to_process.write(cmd)

    def start_record(self, timeout=5):

        ret = super().start_record(timeout=timeout)

        if (self.pose_process is not None) and (self.writer_process is not None):
            if (self.pose_process.is_alive()) and (self.writer_process.is_alive()):
                self.q_to_process.write(("pose", "write", True))

                stime = time.time()
                while time.time() - stime < timeout:
                    cmd = self.q_from_process.read()
                    if cmd is not None:
                        if (cmd[0] == "pose") and (cmd[1] == "write"):
                            ret = cmd[2]
                            break
                        else:
                            self.q_from_process.write(cmd)

        return ret

    def stop_record(self, timeout=5):

        ret = super().stop_record(timeout=timeout)

        if (self.pose_process is not None) and (self.writer_process is not None):
            if (self.pose_process.is_alive()) and (self.writer_process.is_alive()):
                self.q_to_process.write(("pose", "write", False))

                stime = time.time()
                while time.time() - stime < timeout:
                    cmd = self.q_from_process.read()
                    if cmd is not None:
                        if (cmd[0] == "pose") and (cmd[1] == "write"):
                            ret = not cmd[2]
                            break
                        else:
                            self.q_from_process.write(cmd)

        return ret

    def stop_pose_process(self):

        ret = True
        if self.pose_process is not None:
            if self.pose_process.is_alive():
                self.q_to_process.write(("pose", "end"))

                while True:
                    cmd = self.q_from_process.read()
                    if cmd is not None:
                        if cmd[0] == "pose":
                            if cmd[1] == "end":
                                break
                        else:
                            self.q_from_process.write(cmd)

                self.pose_process.join(5)
                if self.pose_process.is_alive():
                    self.pose_process.terminate()

        return True

    def save_pose(self, filename, timeout=60):

        ret = False
        if self.pose_process is not None:
            if self.pose_process.is_alive():
                self.q_to_process.write(("pose", "save", filename))

                stime = time.time()
                while time.time() - stime < timeout:
                    cmd = self.q_from_process.read()
                    if cmd is not None:
                        if (cmd[0] == "pose") and (cmd[1] == "save"):
                            ret = cmd[3]
                            break
                        else:
                            self.q_from_process.write(cmd)
        return ret

    def _save_pose(self, filename):
        """ Saves a pandas data frame with pose data collected while recording video

        Returns
        -------
        bool
            a logical flag indicating whether save was successful
        """

        ret = False

        if len(self.pose_times) > 0:

            dlc_file = f"{filename}_DLC.hdf5"
            proc_file = f"{filename}_PROC"

            bodyparts = self.dlc.cfg["all_joints_names"]
            poses = np.array(self.poses)
            poses = poses.reshape((poses.shape[0], poses.shape[1] * poses.shape[2]))
            pdindex = pd.MultiIndex.from_product(
                [bodyparts, ["x", "y", "likelihood"]], names=["bodyparts", "coords"]
            )
            pose_df = pd.DataFrame(poses, columns=pdindex)
            pose_df["frame_time"] = self.pose_frame_times
            pose_df["pose_time"] = self.pose_times

            pose_df.to_hdf(dlc_file, key="df_with_missing", mode="w")
            if self.dlc.processor is not None:
                self.dlc.processor.save(proc_file)

            self.poses = []
            self.pose_times = []
            self.pose_frame_times = []

            ret = True

        return ret

    def get_display_pose(self):

        pose = self.display_pose_queue.read(clear=True)
        if pose is not None:
            self.display_pose = pose
            if self.device.display_resize != 1:
                self.display_pose[:, :2] *= self.device.display_resize

        return self.display_pose
